import json
import textwrap
from dataclasses import dataclass, field
from pathlib import Path
from test.testutils import Generic, TypeVar
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import pytest
from typing_extensions import Literal

from simple_parsing.helpers import Serializable, dict_field, list_field
from simple_parsing.helpers.serialization.decoding import (
    get_decoding_fn,
    register_decoding_fn,
)
from simple_parsing.helpers.serialization.serializable import loads_json
from simple_parsing.utils import DataclassT


def test_encode_something(simple_attribute):

    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass(Serializable):
        d: Dict[str, some_type] = dict_field()
        l: List[Tuple[some_type, some_type]] = list_field()
        t: Dict[str, Optional[some_type]] = dict_field()
        # w: Dict[str, Union[some_type, int, str, None, str, None]] = dict_field()

    b = SomeClass()
    b.d.update({"hey": expected_value})
    b.l.append((expected_value, expected_value))
    b.t.update({"hey": None, "hey2": expected_value})
    # b.w.update({
    #     "hey": None,
    #     "hey2": "heyo",
    #     "hey3": 1,
    #     "hey4": expected_value,
    # })
    assert SomeClass.loads(b.dumps()) == b


def test_literal_decoding():
    @dataclass
    class SomeClass(Serializable):
        x: Literal["a", "b", "c"] = "a"

    # This test should fail if there's a warning on decoding- previous versions
    # have raised a UserWarning when decoding a literal, of the form:
    # Unable to find a decoding function for annotation typing.Literal['a', 'b', 'c']
    # with pytest.warns(UserWarning, match="Unable to find a decoding function"):
    #     assert SomeClass.loads('{"x": "a"}') == SomeClass()

    # Make sure that we can't decode a value that's not in the literal
    with pytest.raises(TypeError):
        SomeClass.loads('{"x": "d"}')


def test_typevar_decoding(simple_attribute):
    @dataclass
    class Item(Serializable, decode_into_subclasses=True):
        name: str = "chair"
        price: float = 399
        stock: int = 10

    @dataclass
    class DiscountedItem(Item):
        discount_factor: float = 0.5

    ItemT = TypeVar("ItemT", bound=Item)

    @dataclass
    class Container(Serializable, Generic[ItemT]):
        items: List[ItemT] = list_field()

    chair = Item()
    cheap_chair = DiscountedItem(name="Cheap chair")
    c = Container(items=[chair, cheap_chair])

    assert Container.loads(c.dumps()) == c

    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass(Serializable):
        d: Dict[str, some_type] = dict_field()
        l: List[Tuple[some_type, some_type]] = list_field()
        t: Dict[str, Optional[some_type]] = dict_field()
        # w: Dict[str, Union[some_type, int, str, None, str, None]] = dict_field()

    b = SomeClass()
    b.d.update({"hey": expected_value})
    b.l.append((expected_value, expected_value))
    b.t.update({"hey": None, "hey2": expected_value})
    # b.w.update({
    #     "hey": None,
    #     "hey2": "heyo",
    #     "hey3": 1,
    #     "hey4": expected_value,
    # })
    assert SomeClass.loads(b.dumps()) == b


def test_super_nesting():
    @dataclass
    class Complicated(Serializable):
        x: List[List[List[Dict[int, Tuple[int, float, str, List[float]]]]]] = list_field()

    c = Complicated()
    c.x = [[[{0: (2, 1.23, "bob", [1.2, 1.3])}]]]
    assert Complicated.loads(c.dumps()) == c
    assert c.dumps() == '{"x": [[[{"0": [2, 1.23, "bob", [1.2, 1.3]]}]]]}'


@pytest.mark.parametrize(
    "some_type, encoded_value, expected_value",
    [
        # (Tuple[int, float], json.loads(json.dumps([1, 2])), (1, 2.0)),
        (
            List[Tuple[int, float]],
            json.loads(json.dumps([[1, 2], [3, 4]])),
            [(1, 2.0), (3, 4.0)],
        ),
        (Union[int, float], "1", 1),
        (Union[int, float], "1.2", 1.2),
        pytest.param(
            Union[int, float],
            1.2,
            1.2,
            marks=[
                pytest.mark.xfail(reason="decoding an int works (but raises a warning)"),
            ],
        ),
        # NOTE: Here we expect a float, since it's the first type that will work.
        (Union[float, int], "1", 1.0),
        (Union[float, int], "1.2", 1.2),
        (Union[float, int], 1.2, 1.2),
    ],
)
def test_decode(some_type: Type, encoded_value: Any, expected_value: Any):
    decoding_function = get_decoding_fn(some_type)
    actual = decoding_function(encoded_value)
    assert actual == expected_value
    assert type(actual) == type(expected_value)


@dataclass
class Hparams:
    use_log: int = 1
    severity: int = 2
    probs: List[int] = field(default_factory=lambda: [1, 2])


@dataclass
class Parameters(Serializable):
    hparams: Hparams = field(default_factory=Hparams)


def test_implicit_int_casting(tmp_path: Path):
    """Test that we do in fact perform the unsafe casting as described in #227:
    https://github.com/lebrice/SimpleParsing/issues/227
    """
    with open(tmp_path / "conf.yaml", "w") as f:
        f.write(
            textwrap.dedent(
                """\
                hparams:
                    use_log: 1
                    severity: 0.1
                    probs: [0.1, 0.2]
                """
            )
        )
    with pytest.warns(RuntimeWarning, match="Unsafe casting"):
        file_config = Parameters.load(tmp_path / "conf.yaml")
    assert file_config == Parameters(hparams=Hparams(severity=0, probs=[0, 0]))


@pytest.fixture(autouse=True)
def reset_int_decoding_fns_after_test():
    """Reset the decoding function for `int` to the default after each test."""
    from simple_parsing.helpers.serialization.decoding import _decoding_fns

    backup = _decoding_fns.copy()
    yield
    for key, value in _decoding_fns.items():
        if key not in backup:
            # print(f"Test added a decoding function for {key} with value {value}.")
            pass
        elif value != backup[key]:
            # print(
            #     f"Test changed the decoding function for {key} from {backup[key]} to {value}.",
            # )
            pass
    _decoding_fns.clear()
    _decoding_fns.update(backup)


def test_registering_safe_casting_decoding_fn():
    """Test the solution to 'issue' #227: https://github.com/lebrice/SimpleParsing/issues/227"""

    # Solution: register a decoding function for `int` that casts to int, but raises an error if
    # the value would lose precision.

    def _safe_cast(v: Any) -> int:
        int_v = int(v)
        if int_v != float(v):
            raise ValueError(f"Cannot safely cast {v} to int")
        return int_v

    register_decoding_fn(int, _safe_cast, overwrite=True)

    assert (
        Parameters.loads_yaml(
            textwrap.dedent(
                """\
        hparams:
            use_log: 1
            severity: 0.0
            probs: [3, 4.0]
        """
            )
        )
        == Parameters(hparams=Hparams(severity=0, probs=[3, 4]))
    )

    with pytest.raises(ValueError, match="Cannot safely cast 0.1 to int"):
        Parameters.loads_yaml(
            textwrap.dedent(
                """\
            hparams:
                use_log: 1
                severity: 0.1
                probs: [0, 0]
            """
            )
        )

    with pytest.raises(ValueError, match="Cannot safely cast 0.2 to int"):
        Parameters.loads_yaml(
            textwrap.dedent(
                """\
            hparams:
                use_log: 1
                severity: 1
                probs: [0.2, 0.3]
            """
            )
        )


@pytest.mark.xfail(strict=True, match="DID NOT RAISE <class 'ValueError'>")
def test_optional_list_type_doesnt_use_type_decoding_fn():
    """BUG: Parsing an Optional[list[int]] doesn't work correctly."""

    def _safe_cast(v: Any) -> int:
        int_v = int(v)
        if int_v != float(v):
            raise ValueError(f"Cannot safely cast {v} to int")
        return int_v

    register_decoding_fn(int, _safe_cast, overwrite=True)

    with pytest.raises(ValueError):
        get_decoding_fn(List[int])([0.1, 0.2])

    # BUG: This doesn't work correctly.
    with pytest.raises(ValueError):
        get_decoding_fn(Optional[List[int]])([0.1, 0.2])


@dataclass
class ClassWithInt:
    a: int = 1


@dataclass
class ClassWithIntList:
    values: List[int] = field(default_factory=[1, 2, 3].copy)


@pytest.mark.parametrize(
    ("class_to_use", "serialized_dict", "expected_message", "expected_result"),
    [
        pytest.param(
            ClassWithInt,
            {"a": 1.1},
            r"Unsafe casting occurred when deserializing field 'a' of type <class 'int'>: raw value: 1.1, decoded value: 1",
            ClassWithInt(a=int(1.1)),
            id="float to int",
        ),
        pytest.param(
            ClassWithInt,
            {"a": True},
            r"Unsafe casting occurred when deserializing field 'a' of type <class 'int'>: raw value: True, decoded value: 1",
            ClassWithInt(a=int(True)),
            id="bool to int",
        ),
        pytest.param(
            ClassWithIntList,
            {"values": [1.1, 2.2, 3.3]},
            r"Unsafe casting occurred when deserializing field 'values' of type typing.List\[int\]: raw value: \[1.1, 2.2, 3.3\], decoded value: \[1, 2, 3\].",
            ClassWithIntList(values=[int(1.1), int(2.2), int(3.3)]),
            id="List of floats",
        ),
    ],
)
def test_issue_227_unsafe_int_casting_on_load(
    class_to_use: Type[DataclassT],
    serialized_dict: dict,
    expected_message: str,
    expected_result: DataclassT,
):
    """Test that a warning is raised when performing a lossy cast when deserializing a dataclass."""
    with pytest.warns(
        RuntimeWarning,
        match=expected_message,
    ) as record:
        obj = loads_json(class_to_use, json.dumps(serialized_dict))
        assert obj == expected_result
    assert len(record.list) == 1



from __future__ import annotations

import argparse
from typing import Any, Callable, Iterable, Sequence

from typing_extensions import Literal

from .. import utils

DEFAULT_NEGATIVE_PREFIX = "--no"


class BooleanOptionalAction(argparse.Action):
    """Similar to `argparse.BooleanOptionalAction`.

    * Support using a custom negative prefix (makes this compatible with `absl.flags`)
    * Accept `--flag=true` value
    * Support Python 3.8
    """

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: bool | None = None,
        type: Callable[[str], bool] = utils.str2bool,
        choices: Iterable[Any] | None = None,
        required: bool = False,
        help: str | None = None,
        metavar: str | tuple[str, ...] | None = "bool",
        nargs: Literal["?"] | None = "?",
        negative_prefix: str = DEFAULT_NEGATIVE_PREFIX,
        negative_option: str | None = None,
        _conflict_prefix: str | None = "",
    ):
        option_strings = list(option_strings)
        if nargs is None:
            nargs = "?"

        if nargs != "?":
            more_info = ""
            if nargs in {0, 1}:
                more_info = (
                    "In argparse, nargs=0 parses an empty list, and nargs=1 is list of bools "
                    "with one item, not a required single boolean."
                )
            elif nargs in {"+", "*"} or isinstance(nargs, int):
                more_info = (
                    "To parse a field with a list of booleans, use a sequence of booleans as a "
                    "field annotation (e.g. list[bool] or tuple[bool, ...])."
                )

            field = {option.lstrip("-") for option in option_strings}
            raise ValueError(
                f"Invalid nargs for bool field '{'/'.join(field)}': {nargs!r}\n"
                f"Fields with a `bool` annotation only accepts nargs of `'?'` or `None`, since it "
                "parses single-boolean fields. " + "\n" + more_info
            )

        self.negative_prefix = negative_prefix
        self.negative_option = negative_option

        self.negative_option_strings: list[str] = []
        if negative_option is not None:
            # Use the negative option.
            # _conflict_prefix is passed down from the FieldWrapper, and is used to also add a
            # prefix to the generated negative options. This is used to avoid conflicts between
            # the negative options of different fields!
            # For example if both a `train: Config` and `valid: Config` have a `--verbose` flag,
            # with a `--silent` negative option, then we have to add a prefix to the negative flags
            # also!
            after_dashes = ""
            if _conflict_prefix:
                assert _conflict_prefix.endswith(".")
                after_dashes = _conflict_prefix

            if negative_option.startswith("-"):
                negative_option_without_leading_dashes = negative_option.lstrip("-")
                num_leading_dashes = len(negative_option) - len(
                    negative_option_without_leading_dashes
                )

            else:
                negative_option_without_leading_dashes = negative_option
                # NOTE: Pre-emptively changing this here so we don't use a single leading dash when
                # there's prefix.
                # Use a single leading dash only when there isn't a prefix and if the negative
                # option is a single character.
                num_leading_dashes = 2 if len(after_dashes + negative_option) > 1 else 1
            negative_option = (
                "-" * num_leading_dashes + after_dashes + negative_option_without_leading_dashes
            )
            self.negative_option_strings = [negative_option]
        else:
            self.negative_option_strings = []
            for option_string in option_strings:
                if "." in option_string:
                    parts = option_string.split(".")
                    # NOTE: Need to be careful here.
                    first, *middle, last = parts

                    negative_prefix_without_leading_dashes = negative_prefix.lstrip("-")
                    num_leading_dashes = len(negative_prefix) - len(
                        negative_prefix_without_leading_dashes
                    )
                    first_without_leading_dashes = first.lstrip("-")

                    first = "-" * num_leading_dashes + first_without_leading_dashes
                    last = negative_prefix_without_leading_dashes + last

                    negative_option_string = ".".join([first] + middle + [last])
                    self.negative_option_strings.append(negative_option_string)

                elif option_string.startswith("-"):
                    without_leading_dashes = option_string.lstrip("-")
                    negative_option_string = self.negative_prefix + without_leading_dashes
                    if negative_option_string not in self.negative_option_strings:
                        # NOTE: don't want -a and --a to both add a --noa negative option.
                        self.negative_option_strings.append(negative_option_string)
                else:
                    raise NotImplementedError(
                        f"Invalid option string {option_string!r} for boolean field. "
                        f"This action doesn't support positional arguments. "
                        f"Option strings should start with one or more dashes ('-'). "
                    )
        if help is not None and default is not None and default is not argparse.SUPPRESS:
            help += " (default: %(default)s)"

        super().__init__(
            option_strings=option_strings + self.negative_option_strings,
            dest=dest,
            nargs=nargs,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )
        self.type: Callable[[str], bool]
        assert self.type is not None

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ):
        # NOTE: `option_string` is only None when using a positional argument.
        if option_string is None:
            raise NotImplementedError("This action doesn't support positional arguments yet.")
        assert option_string in self.option_strings

        used_negative_flag = option_string in self.negative_option_strings

        bool_value: bool
        if values is None:  # --my_flag / --nomy_flag
            bool_value = not used_negative_flag
        elif used_negative_flag:  # Cannot set `--nomy_flag=True/False`
            parser.exit(
                message=f"Negative flags cannot be passed a value (Got: {option_string}={values})"
            )
        elif isinstance(values, bool):
            bool_value = values
        elif isinstance(values, str):  # --my_flag true
            bool_value = self.type(values)
        else:
            raise ValueError(f"Unsupported value for {option_string}: {values!r}")

        setattr(namespace, self.dest, bool_value)

    def format_usage(self):
        return " | ".join(self.option_strings)



""" Tests for compatibility with the postponed evaluation of annotations. """
from __future__ import annotations

import dataclasses
import sys
import typing
from dataclasses import InitVar, dataclass
from typing import Any, Callable, Generic, TypeVar

import pytest

from simple_parsing import field
from simple_parsing.annotation_utils.get_field_annotations import (
    get_field_type_from_annotations,
)
from simple_parsing.helpers import Serializable
from simple_parsing.utils import is_list, is_tuple

from .testutils import TestSetup


@dataclass
class Foo(TestSetup):
    a: int = 123

    b: str = "fooobar"
    c: tuple[int, float] = (123, 4.56)

    d: list[bool] = field(default_factory=list)

    e: InitVar[int] = 5
    d: int = field(init=False)

    def __post_init__(self, e: int) -> None:
        self.d = e + 2


@dataclass
class Bar(TestSetup):
    barry: Foo = field(default_factory=Foo)
    joe: Foo = field(default_factory=lambda: Foo(b="rrrrr"))
    z: float = 123.456
    some_list: list[float] = field(default_factory=[1.0, 2.0].copy)


def test_future_annotations():
    foo = Foo.setup()
    assert foo == Foo()

    foo = Foo.setup("--a 2 --b heyo --c 1 7.89")
    assert foo == Foo(a=2, b="heyo", c=(1, 7.89))


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 8),
    reason="Before 3.8 `InitVar[tp] is InitVar` so it's impossible to retrieve field type",
)
def test_future_annotations_initvar():
    foo = Foo.setup("--e 6")
    assert foo.d == 8


def test_future_annotations_nested():
    bar = Bar.setup()
    assert bar == Bar()
    assert bar.barry == Foo()
    bar = Bar.setup("--barry.a 2 --barry.b heyo --barry.c 1 7.89")
    assert bar.barry == Foo(a=2, b="heyo", c=(1, 7.89))
    assert isinstance(bar.joe, Foo)


@dataclass
class ClassWithNewUnionSyntax(TestSetup):
    v: int | float = 123


@dataclass
class OtherClassWithNewUnionSyntax(ClassWithNewUnionSyntax):
    """Create a child class without annotations, just to check that they are picked up from the
    base class.
    """


@pytest.mark.parametrize(
    "ClassWithNewUnionSyntax", [ClassWithNewUnionSyntax, OtherClassWithNewUnionSyntax]
)
def test_new_union_syntax(ClassWithNewUnionSyntax: type[ClassWithNewUnionSyntax]):
    assert ClassWithNewUnionSyntax.setup() == ClassWithNewUnionSyntax()
    assert ClassWithNewUnionSyntax.setup("--v 456") == ClassWithNewUnionSyntax(v=456)
    assert ClassWithNewUnionSyntax.setup("--v 4.56") == ClassWithNewUnionSyntax(v=4.56)

    field_annotations = {f.name: f.type for f in dataclasses.fields(ClassWithNewUnionSyntax)}
    from simple_parsing.utils import is_union

    assert is_union(field_annotations["v"])

    # with pytest.raises(Exception):
    #     assert ClassWithNewUnionSyntax.setup("--v bob")


def test_more_complicated_unions():
    """Test that simple-parsing can properly convert the 'new-style' type annotations.

    These are converted to the 'old-style' annotations using the `typing` module, e.g.
    A | B -> Union[A, B]
    tuple[A, B, C] -> Tuple[A,B,C]

    These values are then used to modify the `type` attribute of the `dataclasses.Field` objects
    on the class, *in-place*, so that simple-parsing can work just like before.
    """

    # T = TypeVarTuple("T")  # TODO: Use this eventually (when it becomes possible).
    T = TypeVar("T")
    U = TypeVar("U")

    class Try(Generic[T, U]):
        """Returns a callable that attempts to use the given functions, and returns the first
        result that is obtained without raising an exception.

        If all the functions fail, calls `none_worked` if it's a callable, or returns it as a value
        if it isn't a callable.
        """

        def __init__(
            self,
            *functions: Callable[..., T] | Callable[..., U],
        ) -> None:
            self.functions = functions

        def __str__(self) -> str:
            return (
                f"{type(self).__qualname__}("
                + ", otherwise ".join(f.__qualname__ for f in self.functions)
                + ")"
            )

        def __call__(self, *args: Any, **kwargs: Any) -> T | U:
            exceptions: list[Exception] = []
            for function in self.functions:
                try:
                    return function(*args, **kwargs)
                except Exception as exc:
                    exceptions.append(exc)
            return self.none_worked(exceptions)

        def none_worked(self, exceptions: list[Exception]) -> typing.NoReturn:
            raise RuntimeError(
                "None of the functions worked!\n"
                + "\n".join(
                    "- Function {func} raised: {exc}\n"
                    for func, exc in zip(self.functions, exceptions)
                )
            )

    int_or_float: Try[int, float] = Try(int, float)

    @dataclass
    class MoreComplex(TestSetup):
        vals_list: list[int | float] = field(default_factory=list, type=int_or_float)
        vals_tuple: tuple[int | float, bool] = field(default=(1, False))

    assert (
        get_field_type_from_annotations(MoreComplex, "vals_list")
        == typing.List[typing.Union[int, float]]
    )
    assert MoreComplex.__annotations__["vals_list"] == "list[int | float]"
    assert MoreComplex.__annotations__["vals_tuple"] == "tuple[int | float, bool]"
    assert (
        get_field_type_from_annotations(MoreComplex, "vals_tuple")
        == typing.Tuple[typing.Union[int, float], bool]
    )
    # NOTE: Before we do anything related to simple-parsing, the value in Field.type should still be
    # equivalent to their annotations.
    field_annotations = {f.name: f.type for f in dataclasses.fields(MoreComplex)}
    assert field_annotations["vals_list"] == "list[int | float]"

    # Now, once we add arguments for it, we expect these fields here to have a different `type`,
    # if python < 3.9
    from simple_parsing import ArgumentParser

    ArgumentParser().add_arguments(MoreComplex, dest="unused")
    field_annotations = {f.name: f.type for f in dataclasses.fields(MoreComplex)}
    assert field_annotations["vals_list"] == typing.List[typing.Union[int, float]]
    assert field_annotations["vals_tuple"] == typing.Tuple[typing.Union[int, float], bool]

    assert is_list(field_annotations["vals_list"])
    assert is_tuple(field_annotations["vals_tuple"])

    assert MoreComplex.setup("--vals_list 456 123") == MoreComplex(vals_list=[456, 123])
    assert MoreComplex.setup("--vals_list 4.56 1.23") == MoreComplex(vals_list=[4.56, 1.23])
    # NOTE: Something funky is happening: Seems like the `float` type here is being registered as
    # the handler also in the second case, for the tuple!
    assert MoreComplex.setup("--vals_tuple 456 False") == MoreComplex(vals_tuple=(456, False))
    assert MoreComplex.setup("--vals_tuple 4.56 True") == MoreComplex(vals_tuple=(4.56, True))


@pytest.mark.xfail(reason="TODO: Properly support containers of union types.")
def test_parsing_containers_of_unions():
    @dataclass
    class MoreComplex(TestSetup):
        vals_list: list[int | float] = field(default_factory=list)
        vals_tuple: tuple[int | float, bool] = field(default=(1, False))

    assert MoreComplex.setup("--vals_list 456 123") == MoreComplex(vals_list=[456, 123])
    assert MoreComplex.setup("--vals_list 4.56 1.23") == MoreComplex(vals_list=[4.56, 1.23])
    assert MoreComplex.setup("--vals_tuple 456 False") == MoreComplex(vals_tuple=(456, False))
    assert MoreComplex.setup("--vals_tuple 4.56 True") == MoreComplex(vals_tuple=(4.56, True))


@dataclass
class Opts1(Serializable):
    a: int = 64
    b: float = 1.0


@dataclass
class Opts2(Serializable):
    a: int = 32
    b: float = 0.0


@dataclass
class Wrapper(Serializable):
    opts1: Opts1 = field(default_factory=Opts1)
    opts2: Opts2 = field(default_factory=Opts2)


def test_serialization_deserialization():
    # Show that it's not possible to deserialize nested dataclasses
    opts = Wrapper()
    assert Wrapper in Serializable.subclasses
    assert Opts1 in Serializable.subclasses
    assert Opts2 in Serializable.subclasses
    assert Wrapper.from_dict(opts.to_dict()) == opts
    assert Wrapper.loads_json(opts.dumps_json()) == opts
    assert Wrapper.loads_yaml(opts.dumps_yaml()) == opts


@dataclass
class OptimizerConfig(TestSetup):
    lr_scheduler: str = "cosine"
    """ LR scheduler to use. """


@dataclass
class SubclassOfOptimizerConfig(OptimizerConfig):
    bar: int | float = 123
    """ some dummy arg bar. """


def test_missing_annotation_on_subclass():
    """Test that the annotation are correctly fetched from the base class."""

    assert SubclassOfOptimizerConfig.setup("--lr_scheduler cosine") == SubclassOfOptimizerConfig(
        lr_scheduler="cosine"
    )



""" Tests for the setdefaults method of the parser. """
from dataclasses import dataclass, field
from pathlib import Path

import pytest
import yaml

from simple_parsing.helpers.serialization.serializable import save, to_dict
from simple_parsing.parsing import ArgumentParser
from simple_parsing.wrappers.field_wrapper import NestedMode

from .testutils import TestSetup


@dataclass
class Foo(TestSetup):
    a: int = 123
    b: str = "hello"


def test_set_defaults():
    parser = ArgumentParser()
    parser.add_arguments(Foo, dest="foo")
    parser.set_defaults(foo=Foo(b="HOLA"))
    args = parser.parse_args("")
    assert args.foo == Foo(b="HOLA")


def test_set_defaults_from_file(tmp_path: Path):
    parser = ArgumentParser()
    parser.add_arguments(Foo, dest="foo")

    saved_config = Foo(a=456, b="HOLA")
    config_path = tmp_path / "foo.yaml"
    with open(config_path, "w") as f:
        yaml.dump({"foo": to_dict(saved_config)}, f)

    parser.set_defaults(config_path)
    args = parser.parse_args("")
    assert args.foo == saved_config


def test_set_defaults_from_file_without_root(tmp_path: Path):
    """test that set_defaults accepts the fields of the dataclass directly, when the parser has
    nested_mode=NestedMode.WITHOUT_ROOT.
    """
    parser = ArgumentParser(nested_mode=NestedMode.WITHOUT_ROOT)
    parser.add_arguments(Foo, dest="foo")

    save_path = tmp_path / "temp.json"
    save(dict(a=456, b="BYE BYE"), path=save_path)

    parser.set_defaults(save_path)

    args = parser.parse_args("")
    assert args.foo == Foo(a=456, b="BYE BYE")

    args = parser.parse_args("--a 111".split())
    assert args.foo == Foo(a=111, b="BYE BYE")


def test_set_defaults_from_file_before_adding_args(tmp_path: Path):
    parser = ArgumentParser()

    saved_config = Foo(a=456, b="HOLA")
    config_path = tmp_path / "foo.yaml"
    with open(config_path, "w") as f:
        yaml.dump({"foo": to_dict(saved_config)}, f)
    parser.set_defaults(config_path)

    parser.add_arguments(Foo, dest="foo")
    args = parser.parse_args("")
    assert args.foo == saved_config


@dataclass
class ConfigWithFoo(TestSetup):
    c: str = "bob"
    foo: Foo = field(default_factory=Foo)


@pytest.mark.parametrize("with_root", [True, False])
@pytest.mark.parametrize("add_arguments_before", [True, False])
def test_with_nested_field(tmp_path: Path, add_arguments_before: bool, with_root: bool):
    """Test that when we use set_defaults with a config that has a nested dataclass field,
    we can pass a path to a yaml file for one of the field, and it also works.
    """
    parser = ArgumentParser(
        nested_mode=NestedMode.WITHOUT_ROOT if not with_root else NestedMode.DEFAULT
    )
    if add_arguments_before:
        parser.add_arguments(ConfigWithFoo, dest="config")

    save_path = tmp_path / "temp.json"
    from simple_parsing.helpers.serialization import encode

    saved_config = ConfigWithFoo(foo=Foo(a=456, b="BYE BYE"))

    if with_root:
        save(encode({"config": saved_config}), path=save_path)
    else:
        save(saved_config, path=save_path)
    parser.set_defaults(save_path)

    if not add_arguments_before:
        parser.add_arguments(ConfigWithFoo, dest="config")

    args = parser.parse_args("")
    assert args.config == saved_config

    args = parser.parse_args("--a 111".split())
    assert args.config == ConfigWithFoo(foo=Foo(a=111, b="BYE BYE"))



""" Tests for issue 144: https://github.com/lebrice/SimpleParsing/issues/144 """
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Union

import pytest

from simple_parsing.helpers.serialization.serializable import Serializable


class TestOptional:
    @dataclass
    class Foo(Serializable):
        foo: Optional[int] = 123

    @pytest.mark.parametrize("d", [{"foo": None}, {"foo": 1}])
    def test_round_trip(self, d: dict):
        # NOTE: this double round-trip makes the comparison agnostic to any conversion that may
        # happen between the raw dict values and the arguments of the dataclasses.
        assert self.Foo.from_dict(self.Foo.from_dict(d).to_dict()) == self.Foo.from_dict(d)


class TestUnion:
    @dataclass
    class Foo(Serializable):
        foo: Union[int, dict[int, bool]] = 123

    @pytest.mark.parametrize("d", [{"foo": None}, {"foo": {1: "False"}}])
    def test_round_trip(self, d: dict):
        # NOTE: this double round-trip makes the comparison agnostic to any conversion that may
        # happen between the raw dict values and the arguments of the dataclasses.
        assert self.Foo.from_dict(self.Foo.from_dict(d).to_dict()) == self.Foo.from_dict(d)


class TestList:
    @dataclass
    class Foo(Serializable):
        foo: List[int] = field(default_factory=list)

    @pytest.mark.parametrize("d", [{"foo": []}, {"foo": [123, 456]}])
    def test_round_trip(self, d: dict):
        # NOTE: this double round-trip makes the comparison agnostic to any conversion that may
        # happen between the raw dict values and the arguments of the dataclasses.
        assert self.Foo.from_dict(self.Foo.from_dict(d).to_dict()) == self.Foo.from_dict(d)


class TestTuple:
    @dataclass
    class Foo(Serializable):
        foo: Tuple[int, float, bool]

    @pytest.mark.parametrize("d", [{"foo": (1, 1.2, False)}, {"foo": ("1", "1.2", "True")}])
    def test_round_trip(self, d: dict):
        # NOTE: this double round-trip makes the comparison agnostic to any conversion that may
        # happen between the raw dict values and the arguments of the dataclasses.
        assert self.Foo.from_dict(self.Foo.from_dict(d).to_dict()) == self.Foo.from_dict(d)


class TestDict:
    @dataclass
    class Foo(Serializable):
        foo: Dict[int, float] = field(default_factory=dict)

    @pytest.mark.parametrize("d", [{"foo": {}}, {"foo": {"123": "4.56"}}])
    def test_round_trip(self, d: dict):
        # NOTE: this double round-trip makes the comparison agnostic to any conversion that may
        # happen between the raw dict values and the arguments of the dataclasses.
        assert self.Foo.from_dict(self.Foo.from_dict(d).to_dict()) == self.Foo.from_dict(d)



from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional, Tuple, Type

import pytest

from simple_parsing import ArgumentParser, ConflictResolution, field
from simple_parsing.utils import str2bool
from simple_parsing.wrappers.field_parsing import parse_enum


def test_cmd_false_doesnt_create_conflicts():
    @dataclass
    class A:
        batch_size: int = field(default=10, cmd=False)

    @dataclass
    class B:
        batch_size: int = 20

    # @dataclass
    # class Foo(TestSetup):
    #     a: A = mutable_field(A)
    #     b: B = mutable_field(B)

    parser = ArgumentParser(conflict_resolution=ConflictResolution.NONE)
    parser.add_arguments(A, "a")
    parser.add_arguments(B, "b")
    args = parser.parse_args("--batch_size 32".split())
    a: A = args.a
    b: B = args.b
    assert a == A()
    assert b == B(batch_size=32)


class Color(Enum):
    blue: str = "BLUE"
    red: str = "RED"
    green: str = "GREEN"
    orange: str = "ORANGE"


@pytest.mark.xfail(
    reason="Removed this function. TODO: see https://github.com/lebrice/SimpleParsing/issues/150."
)
@pytest.mark.parametrize(
    "annotation, expected_options",
    [
        (Tuple[int, int], dict(nargs=2, type=int)),
        (Tuple[Color, Color], dict(nargs=2, type=parse_enum(Color))),
        (
            Optional[Tuple[Color, Color]],
            dict(nargs=2, type=parse_enum(Color), required=False),
        ),
        (List[str], dict(nargs="*", type=str)),
        (Optional[List[str]], dict(nargs="*", type=str, required=False)),
        (Optional[str], dict(nargs="?", type=str, required=False)),
        (Optional[bool], dict(nargs="?", type=str2bool, required=False)),
        # (Optional[Tuple[Color, str]], dict(nargs=2, type=get_parsing_fn(Tuple[Color, str]), required=False)),
    ],
)
def test_generated_options_from_annotation(annotation: Type, expected_options: Dict):
    raise NotImplementedError(
        """
        TODO: Would be a good idea to refactor the FieldWrapper class a bit. The args_dict (a dict
        of all the argparse arguments for a given field, that get passed to parser.add_arguments
        in the FieldWrapper) is currently created using a mix of three things (with increasing
        priority):
        - The type annotation
        - The dataclass context (e.g. when adding an Optional[Dataclass] field on another
          dataclass, or when using the `default` or `prefix` arguments to `parser.add_arguments`.
        - The manual overrides (arguments of `parser.add_argument` passed to the `field` function)

        These three are currently a bit mixed together in the `FieldWrapper` class. It would be
        preferable to design a way for them to be cleanly separated.
        """
    )
    # from simple_parsing.wrappers.field_wrapper import get_argparse_options_for_annotation

    # actual_options = get_argparse_options_for_annotation(annotation)
    # for option, expected_value in expected_options.items():
    #     assert actual_options[option] == expected_value



from __future__ import annotations

import argparse
import dataclasses
import functools
import inspect
import sys
import textwrap
from dataclasses import MISSING
from logging import getLogger
from typing import Any, Callable, Generic, TypeVar, cast

import docstring_parser as dp
from typing_extensions import Literal

from .. import docstring, utils
from ..utils import Dataclass, DataclassT, is_dataclass_instance, is_dataclass_type
from .field_wrapper import FieldWrapper
from .wrapper import Wrapper

logger = getLogger(__name__)

MAX_DOCSTRING_DESC_LINES_HEIGHT: int = 50
"""
Maximum number of lines of the class docstring to include in the autogenerated argument group
description. If fields don't have docstrings or help text, then this is not used, and the entire
docstring is used as the description of the argument group.
"""

DataclassWrapperType = TypeVar("DataclassWrapperType", bound="DataclassWrapper")


class DataclassWrapper(Wrapper, Generic[DataclassT]):
    def __init__(
        self,
        dataclass: type[DataclassT],
        name: str,
        default: DataclassT | dict | None = None,
        prefix: str = "",
        parent: DataclassWrapper | None = None,
        _field: dataclasses.Field | None = None,
        field_wrapper_class: type[FieldWrapper] = FieldWrapper,
        dataclass_fn: Callable[..., DataclassT] | None = None,
    ):
        super().__init__()
        self.dataclass = dataclass
        self._name = name
        assert is_dataclass_type(dataclass)  # FIXME: Remove
        if dataclass_fn:
            assert callable(dataclass_fn), dataclass_fn
        self.dataclass_fn = dataclass_fn or dataclass
        assert not is_dataclass_instance(self.dataclass_fn)  # FIXME: Remove

        self._default = default
        self.prefix = prefix
        self._parent = parent
        # the field of the parent, which contains this child dataclass.
        self._field = _field
        self.field_wrapper_class = field_wrapper_class

        self.fields: list[FieldWrapper] = []
        self.optional: bool = False

        self._destinations: list[str] = []
        self._required: bool = False
        self._explicit: bool = False
        self._dest: str = ""
        self._children: list[DataclassWrapper] = []
        # the default value(s).
        # NOTE: This is a list only because of the `ConflictResolution.ALWAYS_MERGE` option.
        self._defaults: list[DataclassT] = [default] if default else []

        dataclass_fields: tuple[dataclasses.Field, ...] = _get_dataclass_fields(dataclass)
        # Create an object for each field, which is used to compute (and hold) the arguments that
        # will then be passed to `argument_group.add_argument` later.
        # This could eventually be refactored into a stateless thing. But for now it isn't.
        for field in dataclass_fields:
            if not field.init or field.metadata.get("cmd", True) is False:
                # Don't add arguments for this field.
                continue

            if isinstance(field.type, str):
                # NOTE: Here we'd like to convert the fields type to an actual type, in case the
                # `from __future__ import annotations` feature is used.
                from simple_parsing.annotation_utils.get_field_annotations import (
                    get_field_type_from_annotations,
                )

                field_type = get_field_type_from_annotations(self.dataclass, field.name)
                # Modify the `type` of the Field object, in-place.
                field.type = field_type
            else:
                field_type = field.type

            # Manually overwrite the field default value with the corresponding attribute of the
            # default for the parent.
            field_default = dataclasses.MISSING
            if isinstance(dataclass_fn, functools.partial) and field.name in dataclass_fn.keywords:
                # NOTE: We need to override the default value of the field, because since the
                # dataclass_fn is a partial, and we always set the defaults for all fields in the
                # constructor arguments dict, those would be passed to the partial, and the value
                # for that argument in the partial (e.g. `dataclass_fn = partial(A, a=123)`) would
                # be unused when we call `dataclass_fn(**constructor_args[dataclass_dest])` later.
                field_default = dataclass_fn.keywords[field.name]
                # TODO: This is currently only really necessary in the case where the dataclass_fn
                # is a `functools.partial` (e.g. when using subgroups). But the idea of specifying
                # the default value and passing it here to the wrapper, rather than have the
                # wrappers "fetch" it from their field or their parent, makes sense!
                logger.debug(
                    f"Got a default value of {field_default} for field {field.name} from "
                    f"inspecting the dataclass function! ({dataclass_fn})"
                )
            elif isinstance(default, dict):
                if field.name in default:
                    field_default = default[field.name]
            elif default not in (None, argparse.SUPPRESS):
                field_default = getattr(default, field.name)

            if utils.is_tuple_or_list_of_dataclasses(field_type):
                raise NotImplementedError(
                    f"Field {field.name} is of type {field_type}, which isn't "
                    f"supported yet. (container of a dataclass type)"
                )

            if utils.is_subparser_field(field) or utils.is_choice(field):
                field_wrapper = self.field_wrapper_class(
                    field,
                    parent=self,
                    prefix=prefix,
                )
                if field_default is not dataclasses.MISSING:
                    field_wrapper.set_default(field_default)

                self.fields.append(field_wrapper)

            elif dataclasses.is_dataclass(field_type) and field.default is not None:
                # Non-optional dataclass field.
                # handle a nested dataclass attribute
                dataclass, name = field_type, field.name
                # todo: Figure out if this is still necessary, or if `field_default` can be handled
                # the same way as above.
                if field_default is dataclasses.MISSING:
                    field_default = None
                child_wrapper = DataclassWrapper(
                    dataclass,
                    name,
                    parent=self,
                    _field=field,
                    default=field_default,
                )
                self._children.append(child_wrapper)

            elif utils.contains_dataclass_type_arg(field_type):
                # Extract the dataclass type from the annotation of the field.
                field_dataclass = utils.get_dataclass_type_arg(field_type)
                # todo: Figure out if this is still necessary, or if `field_default` can be handled
                # the same way as above.
                if field_default is dataclasses.MISSING:
                    field_default = None
                child_wrapper = DataclassWrapper(
                    field_dataclass,
                    name=field.name,
                    parent=self,
                    _field=field,
                    default=field_default,
                )
                child_wrapper.required = False
                child_wrapper.optional = True
                self._children.append(child_wrapper)

            else:
                # a "normal" attribute
                field_wrapper = self.field_wrapper_class(field, parent=self, prefix=self.prefix)
                logger.debug(
                    f"wrapped field at {field_wrapper.dest} has a default value of {field_wrapper.default}"
                )
                if field_default is not dataclasses.MISSING:
                    field_wrapper.set_default(field_default)

                self.fields.append(field_wrapper)

        logger.debug(f"The dataclass at attribute {self.dest} has default values: {self.defaults}")

    def add_arguments(self, parser: argparse.ArgumentParser):
        from ..parsing import ArgumentParser

        parser = cast(ArgumentParser, parser)

        group = parser.add_argument_group(title=self.title, description=self.description)

        for wrapped_field in self.fields:
            # Note: This should be true since we don't create a FieldWrapper for fields with
            # `cmd=False`.
            assert wrapped_field.field.metadata.get("cmd", True)

            if wrapped_field.is_subparser:
                wrapped_field.add_subparsers(parser)
                continue

            arg_options = wrapped_field.arg_options

            if argparse.SUPPRESS in self.defaults:
                arg_options["default"] = argparse.SUPPRESS
            if wrapped_field.is_subgroup:
                # NOTE: Not skipping subgroup fields, because even though they will have been
                # resolved at this point, we still want them to show up in the --help message!
                logger.debug(
                    f"Adding a subgroup field {wrapped_field.name} just so it shows up in the "
                    f"--help text."
                )

            logger.info(f"group.add_argument(*{wrapped_field.option_strings}, **{arg_options})")
            # TODO: Perhaps we could hook into the `action` that is returned here to know if the
            # flag was passed or not for a given field.
            _ = group.add_argument(*wrapped_field.option_strings, **arg_options)

    def equivalent_argparse_code(self, leading="group") -> str:
        code = ""
        code += textwrap.dedent(
            f"""
        group = parser.add_argument_group(title="{self.title.strip()}", description="{self.description.strip()}")
        """
        )
        for wrapped_field in self.fields:
            if wrapped_field.is_subparser:
                # TODO:
                raise NotImplementedError("Subparsers equivalent is TODO.")
                code += textwrap.dedent(
                    f"""\
                # add subparsers for each dataclass type in the field.
                subparsers = parser.add_subparsers(
                    title={wrapped_field.name},
                    description={wrapped_field.help},
                    dest={wrapped_field.dest},
                )
                subparsers.required = True

                for subcommand, dataclass_type in {self.subparsers_dict.items()}:
                    subparser = subparsers.add_parser(subcommand)
                    subparser = cast(ArgumentParser, subparser)
                    subparser.add_arguments(dataclass_type, dest=self.dest)
                """
                )
            elif wrapped_field.arg_options:
                code += textwrap.dedent(wrapped_field.equivalent_argparse_code()) + "\n"
        return code

    @property
    def name(self) -> str:
        return self._name

    @property
    def parent(self) -> DataclassWrapper | None:
        return self._parent

    @property
    def defaults(self) -> list[DataclassT | dict[str, Any] | None | Literal[argparse.SUPPRESS]]:
        if self._defaults:
            return self._defaults
        if self._field is None:
            return []
        assert self.parent is not None
        if self.parent.defaults:
            self._defaults = []
            for default in self.parent.defaults:
                if default not in (None, argparse.SUPPRESS):
                    default = getattr(default, self.name)
                self._defaults.append(default)
        else:
            default_field_value = utils.default_value(self._field)
            if default_field_value is MISSING:
                self._defaults = []
            else:
                self._defaults = [default_field_value]
        return self._defaults

    @defaults.setter
    def defaults(self, value: list[DataclassT]):
        self._defaults = value

    @property
    def default(self) -> DataclassT | None:
        return self._default

    # @default.setter
    # def default(self, value: DataclassT) -> None:
    #     self._default = value

    def set_default(self, value: DataclassT | dict | None):
        """Sets the default values for the arguments of the fields of this dataclass."""
        if value is not None and not isinstance(value, dict):
            field_default_values = dataclasses.asdict(value)
        else:
            field_default_values = value
        self._default = value
        if field_default_values is None:
            return
        for field_wrapper in self.fields:
            if field_wrapper.name not in field_default_values:
                continue
            # Manually set the default value for this argument.
            field_default_value = field_default_values[field_wrapper.name]
            field_wrapper.set_default(field_default_value)
        for nested_dataclass_wrapper in self._children:
            if nested_dataclass_wrapper.name not in field_default_values:
                continue
            field_default_value = field_default_values[nested_dataclass_wrapper.name]
            nested_dataclass_wrapper.set_default(field_default_value)

    @property
    def title(self) -> str:
        names_string = f""" [{', '.join(f"'{dest}'" for dest in self.destinations)}]"""
        title = self.dataclass.__qualname__ + names_string
        return title

    @property
    def description(self) -> str:
        if self.parent and self._field:
            doc = docstring.get_attribute_docstring(self.parent.dataclass, self._field.name)
            if doc is not None:
                if doc.docstring_below:
                    return doc.docstring_below
                elif doc.comment_above:
                    return doc.comment_above
                elif doc.comment_inline:
                    return doc.comment_inline

        # NOTE: The class docstring may be EXTRELEMY LARGE.

        class_docstring = inspect.getdoc(self.dataclass) or ""
        if not class_docstring:
            return ""

        doc = dp.parse(class_docstring)

        from simple_parsing.decorators import _description_from_docstring

        description = _description_from_docstring(doc)

        num_lines = len(description.splitlines())
        shortened_description = (
            "\n".join(description.splitlines()[:MAX_DOCSTRING_DESC_LINES_HEIGHT]) + " ..."
        )

        fields_have_docstrings = any(f._docstring.help_string for f in self.fields)
        docstring_is_huge = num_lines > MAX_DOCSTRING_DESC_LINES_HEIGHT
        if not fields_have_docstrings:
            # The fields don't have docstrings. Return the entire docstring, regardless of its
            # size.
            return description
        # Fields have docstrings, so there's probably some duplication between the docstring and
        # the dataclass fields help. Shorten the docstring, if needed.
        if docstring_is_huge:
            return shortened_description
        return description

    # @property
    # def prefix(self) -> str:
    #     return self._prefix

    # @prefix.setter
    # def prefix(self, value: str):
    #     self._prefix = value
    #     for child_wrapper in self._children:
    #         child_wrapper.prefix = value

    @property
    def required(self) -> bool:
        return self._required

    @required.setter
    def required(self, value: bool):
        self._required = value
        for field in self.fields:
            field.required = value
        for child_wrapper in self._children:
            child_wrapper.required = value

    @property
    def multiple(self) -> bool:
        return len(self.destinations) > 1

    @property
    def descendants(self):
        for child in self._children:
            yield child
            yield from child.descendants

    @property
    def dest(self):
        lineage = []
        parent = self.parent
        while parent is not None:
            lineage.append(parent.name)
            parent = parent.parent
        lineage = list(reversed(lineage))
        lineage.append(self.name)
        _dest = ".".join(lineage)
        # logger.debug(f"getting dest, returning {_dest}")
        return _dest

    @property
    def destinations(self) -> list[str]:
        if not self._destinations:
            if self.parent:
                self._destinations = [f"{d}.{self.name}" for d in self.parent.destinations]
            else:
                self._destinations = [self.name]
        return self._destinations

    @destinations.setter
    def destinations(self, value: list[str]):
        self._destinations = value

    def merge(self, other: DataclassWrapper):
        """Absorb all the relevant attributes from another wrapper.
        Args:
            other (DataclassWrapper): Another instance to absorb into this one.
        """
        # logger.debug(f"merging \n{self}\n with \n{other}")
        logger.debug(f"self destinations: {self.destinations}")
        logger.debug(f"other destinations: {other.destinations}")
        # assert not set(self.destinations).intersection(set(other.destinations)), "shouldn't have overlap in destinations"
        # self.destinations.extend(other.destinations)
        for dest in other.destinations:
            if dest not in self.destinations:
                self.destinations.append(dest)
        logger.debug(f"destinations after merge: {self.destinations}")
        self.defaults.extend(other.defaults)

        # Unset the default value for all fields.
        # TODO: Shouldn't be needed anymore.
        for field_wrapper in self.fields:
            field_wrapper.set_default(None)

        for child, other_child in zip(self._children, other._children):
            child.merge(other_child)


def _get_dataclass_fields(dataclass: type[Dataclass]) -> tuple[dataclasses.Field, ...]:
    # NOTE: `dataclasses.fields` method retrieves only `dataclasses._FIELD`
    # NOTE: but we also want to know about `dataclasses._FIELD_INITVAR`
    # NOTE: therefore we partly copy-paste its implementation
    if sys.version_info[:2] < (3, 8):
        # Before 3.8 `InitVar[tp] is InitVar` so it's impossible to retrieve field type
        # therefore we should skip it just to be fully backward compatible
        return dataclasses.fields(dataclass)
    try:
        dataclass_fields_map = getattr(dataclass, dataclasses._FIELDS)
    except AttributeError:
        raise TypeError("must be called with a dataclass type or instance")
    return tuple(
        field
        for field in dataclass_fields_map.values()
        if field._field_type in (dataclasses._FIELD, dataclasses._FIELD_INITVAR)
    )



""" Utility functions that simplify defining field of dataclasses.
"""
from __future__ import annotations

import dataclasses
import functools
import inspect
import warnings
from collections import OrderedDict
from dataclasses import _MISSING_TYPE, MISSING
from enum import Enum
from logging import getLogger
from typing import Any, Callable, Hashable, Iterable, TypeVar, overload

from typing_extensions import Literal, ParamSpec

from simple_parsing.helpers.custom_actions import (
    DEFAULT_NEGATIVE_PREFIX,
    BooleanOptionalAction,
)
from simple_parsing.utils import DataclassT, str2bool

# NOTE: backward-compatibility import because it was moved to a different file.
from .subgroups import subgroups  # noqa: F401

logger = getLogger(__name__)

E = TypeVar("E", bound=Enum)
K = TypeVar("K", bound=Hashable)
V = TypeVar("V")
T = TypeVar("T")


def field(
    default: T | _MISSING_TYPE = MISSING,
    alias: str | list[str] | None = None,
    cmd: bool = True,
    positional: bool = False,
    *,
    to_dict: bool = True,
    encoding_fn: Callable[[T], Any] | None = None,
    decoding_fn: Callable[[Any], T] | None = None,
    # dataclasses.field arguments
    default_factory: Callable[[], T] | _MISSING_TYPE = MISSING,
    init: bool = True,
    repr: bool = True,
    hash: bool | None = None,
    compare: bool = True,
    metadata: dict[str, Any] | None = None,
    **custom_argparse_args: Any,
) -> T:
    """Extension of the `dataclasses.field` function.

    Adds the ability to customize how this field's command-line options are
    created, as well as how it is serialized / deseralized (if the containing
    dataclass inherits from `simple_parsing.Serializable`.

    Leftover arguments are fed directly to the
    `ArgumentParser.add_argument(*option_strings, **kwargs)` method.

    Parameters
    ----------
    default : Union[T, _MISSING_TYPE], optional
        The default field value (same as in `dataclasses.field`), by default MISSING
    alias : Union[str, List[str]], optional
        Additional option_strings to pass to the `add_argument` method, by
        default None. When passing strings which do not start by "-" or "--",
        will be prefixed with "-" if the string is one character and by "--"
        otherwise.
    cmd: bool, optional
        Whether to add command-line arguments for this field or not. Defaults to
        True.

    ## Serialization-related Keyword Arguments:

    to_dict : bool
        Whether to include this field in the dictionary when calling `to_dict()`.
        Defaults to True.
        Only has an effect when the dataclass containing this field is
        `Serializable`.
    encoding_fn : Callable[[T], Any], optional
        Function to apply to this field's value when encoding the dataclass to a
        dict. Only has an effect when the dataclass containing this field is
        `Serializable`.
    decoding_fn : Callable[[Any], T]. optional
        Function to use in order to recover a the value of this field from a
        serialized entry in a dictionary (inside `cls.from_dict`).
        Only has an effect when the dataclass containing this field is
        `Serializable`.

    ## Keyword Arguments of `dataclasses.field`

    default_factory : Union[Callable[[], T], _MISSING_TYPE], optional
        (same as in `dataclasses.field`), by default None
    init : bool, optional
        (same as in `dataclasses.field`), by default True
    repr : bool, optional
        (same as in `dataclasses.field`), by default True
    hash : bool, optional
        (same as in `dataclasses.field`), by default None
    compare : bool, optional
        (same as in `dataclasses.field`), by default True
    metadata : Dict[str, Any], optional
        (same as in `dataclasses.field`), by default None

    Returns
    -------
    T
        The value returned by the `dataclasses.field` function.
    """
    _metadata: dict[str, Any] = metadata if metadata is not None else {}
    if alias:
        _metadata["alias"] = alias if isinstance(alias, list) else [alias]
    _metadata.update(dict(to_dict=to_dict))
    if encoding_fn is not None:
        _metadata.update(dict(encoding_fn=encoding_fn))
    if decoding_fn is not None:
        _metadata.update(dict(decoding_fn=decoding_fn))
    _metadata["cmd"] = cmd
    _metadata["positional"] = positional

    if custom_argparse_args:
        _metadata.update({"custom_args": custom_argparse_args})

        action = custom_argparse_args.get("action")
        if action == "store_false":
            if default not in {MISSING, True}:
                raise RuntimeError(
                    "default should either not be passed or set "
                    "to True when using the store_false action."
                )
            default = True  # type: ignore

        elif action == "store_true":
            if default not in {MISSING, False}:
                raise RuntimeError(
                    "default should either not be passed or set "
                    "to False when using the store_true action."
                )
            default = False  # type: ignore

    if default is not MISSING:
        return dataclasses.field(  # type: ignore
            default=default,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata,
        )
    elif not isinstance(default_factory, dataclasses._MISSING_TYPE):
        return dataclasses.field(
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata,
        )
    else:
        return dataclasses.field(
            init=init, repr=repr, hash=hash, compare=compare, metadata=_metadata
        )


@overload
def choice(
    choices: type[E],
    *,
    default: E,
    default_factory: Callable[[], E] | _MISSING_TYPE = MISSING,
    **kwargs,
) -> E:
    ...


@overload
def choice(choices: dict[K, V], *, default: K, **kwargs) -> V:
    ...


@overload
def choice(
    *choices: T,
    default: T | _MISSING_TYPE = MISSING,
    default_factory: Callable[[], T] | _MISSING_TYPE = MISSING,
    **kwargs,
) -> T:
    ...


def choice(*choices, default=MISSING, **kwargs):
    """Makes a field which can be chosen from the set of choices from the
    command-line.

    Returns a regular `dataclasses.field()`, but with metadata which indicates
    the allowed values.

    (New:) If `choices` is a dictionary, then passing the 'key' will result in
    the corresponding value being used. The values may be objects, for example.
    Similarly for Enum types, passing a type of enum will

    Args:
        default (T, optional): The default value of the field. Defaults to dataclasses.MISSING,
        in which case the command-line argument is required.

    Raises:
        ValueError: If the default value isn't part of the given choices.

    Returns:
        T: the result of the usual `dataclasses.field()` function (a dataclass field/attribute).
    """
    assert len(choices) > 0, "Choice requires at least one positional argument!"

    if len(choices) == 1:
        choices = choices[0]
        if inspect.isclass(choices) and issubclass(choices, Enum):
            # If given an enum, construct a mapping from names to values.
            choice_enum: type[Enum] = choices
            choices = OrderedDict((e.name, e) for e in choice_enum)
            if default is not MISSING and not isinstance(default, choice_enum):
                if default in choices:
                    warnings.warn(
                        UserWarning(
                            f"Setting default={default} could perhaps be ambiguous "
                            f"(enum names vs enum values). Consider using the enum "
                            f"value {choices[default]} instead."
                        )
                    )
                    default = choices[default]
                else:
                    raise ValueError(
                        f"'default' arg should be of type {choice_enum}, but got {default}"
                    )

        if isinstance(choices, dict):
            # if the choices is a dict, the options are the keys
            # save the info about the choice_dict in the field metadata.
            metadata = kwargs.setdefault("metadata", {})
            choice_dict = choices
            # save the choice_dict in metadata so that we can recover the values in postprocessing.
            metadata["choice_dict"] = choice_dict
            choices = list(choice_dict.keys())

            # TODO: If the choice dict is given, then add encoding/decoding functions that just
            # get/set the right key.
            def _encoding_fn(value: Any) -> str:
                """Custom encoding function that will simply represent the value as the
                the key in the dict rather than the value itself.
                """
                if value in choice_dict.keys():
                    return value
                elif value in choice_dict.values():
                    return [k for k, v in choice_dict.items() if v == value][0]
                return value

            kwargs.setdefault("encoding_fn", _encoding_fn)

            def _decoding_fn(value: Any) -> Any:
                """Custom decoding function that will retrieve the value from the
                stored key in the dictionary.
                """
                return choice_dict.get(value, value)

            kwargs.setdefault("decoding_fn", _decoding_fn)

    return field(default=default, choices=choices, **kwargs)


def list_field(*default_items: T, **kwargs) -> list[T]:
    """shorthand function for setting a `list` attribute on a dataclass,
    so that every instance of the dataclass doesn't share the same list.

    Accepts any of the arguments of the `dataclasses.field` function.

    Returns:
        List[T]: a `dataclasses.field` of type `list`, containing the `default_items`.
    """
    if "default" in kwargs and isinstance(kwargs["default"], list):
        assert not default_items
        # can't have that. field wants a default_factory.
        # we just give back a copy of the list as a default factory,
        # but this should be discouraged.
        from copy import deepcopy

        default_factory = functools.partial(deepcopy, kwargs.pop("default"))
    else:
        default_factory = functools.partial(list, default_items)

    return field(default_factory=default_factory, **kwargs)


def dict_field(default_items: dict[K, V] | Iterable[tuple[K, V]] = (), **kwargs) -> dict[K, V]:
    """shorthand function for setting a `dict` attribute on a dataclass,
    so that every instance of the dataclass doesn't share the same `dict`.

    NOTE: Do not use keyword arguments as you usually would with a dictionary
    (as in something like `dict_field(a=1, b=2, c=3)`). Instead pass in a
    dictionary instance with the items: `dict_field(dict(a=1, b=2, c=3))`.
    The reason for this is that the keyword arguments are interpreted as custom
    argparse arguments, rather than arguments of the `dict` function!)

    Also accepts any of the arguments of the `dataclasses.field` function.

    Returns:
        Dict[K, V]: a `dataclasses.Field` of type `Dict[K, V]`, containing the `default_items`.
    """
    return field(default_factory=functools.partial(dict, default_items), **kwargs)


def set_field(*default_items: T, **kwargs) -> set[T]:
    return field(default_factory=functools.partial(set, default_items), **kwargs)


P = ParamSpec("P")


def mutable_field(
    fn: Callable[P, T],
    init: bool = True,
    repr: bool = True,
    hash: bool | None = None,
    compare: bool = True,
    metadata: dict[str, Any] | None = None,
    *fn_args: P.args,
    **fn_kwargs: P.kwargs,
) -> T:
    """Shorthand for `dataclasses.field(default_factory=functools.partial(fn, *fn_args, **fn_kwargs))`.

    NOTE: The *fn_args and **fn_kwargs here are passed to `fn`, and are never used by the argparse
    Action!
    """
    # TODO: Use this 'smart' partial to make it easier to define nested fields.
    # from simple_parsing.helpers.nested_partial import npartial
    default_factory = functools.partial(fn, *fn_args, **fn_kwargs)
    return dataclasses.field(
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


def subparsers(
    subcommands: dict[str, type[DataclassT]],
    default: DataclassT | _MISSING_TYPE = MISSING,
    **kwargs,
) -> Any:
    return field(
        metadata={
            "subparsers": subcommands,
        },
        default=default,
        **kwargs,
    )


@overload
def flag(
    default: _MISSING_TYPE = MISSING,
    *,
    default_factory: _MISSING_TYPE = MISSING,
    negative_prefix: str | None = DEFAULT_NEGATIVE_PREFIX,
    negative_option: str | None = None,
    nargs: Literal["?"] | None = None,
    type: Callable[[str], bool] = str2bool,
    action: type[BooleanOptionalAction] = BooleanOptionalAction,
    **kwargs,
) -> bool:
    ...


@overload
def flag(
    default: bool,
    *,
    default_factory: _MISSING_TYPE = MISSING,
    negative_prefix: str | None = DEFAULT_NEGATIVE_PREFIX,
    negative_option: str | None = None,
    nargs: Literal["?"] | None = None,
    type: Callable[[str], bool] = str2bool,
    action: type[BooleanOptionalAction] = BooleanOptionalAction,
    **kwargs,
) -> bool:
    ...


@overload
def flag(
    default: _MISSING_TYPE = MISSING,
    *,
    default_factory: Callable[[], bool] = ...,
    negative_prefix: str | None = DEFAULT_NEGATIVE_PREFIX,
    negative_option: str | None = None,
    nargs: Literal["?"] | None = None,
    type: Callable[[str], bool] = str2bool,
    action: type[BooleanOptionalAction] = BooleanOptionalAction,
    **kwargs,
) -> bool:
    ...


def flag(
    default: bool | _MISSING_TYPE = MISSING,
    *,
    default_factory: Callable[[], bool] | _MISSING_TYPE = MISSING,
    negative_prefix: str | None = DEFAULT_NEGATIVE_PREFIX,
    negative_option: str | None = None,
    nargs: Literal["?"] | None = None,
    type: Callable[[str], bool] = str2bool,
    action: type[BooleanOptionalAction] = BooleanOptionalAction,
    **kwargs,
) -> bool:
    """A boolean field with a positive and negative command-line argument.

    If either `default` or `default_factory` are set, then both the field and the generated
    command-line arguments are optional. Otherwise, both are required.

    Negative flags are generated using `negative_prefix` and `negative_option`:
    - When `negative_option` is passed, it is used to create the negative flag.
    - Otherwise, `negative_prefix` is prepended to the field name to create the negative flag.

    NOTE: The negative flags don't accept a value. (i.e. `--noverbose` works, but
    `--noverbose=True` does not.)
    The positive flags can be used either with or without a value.
    """
    return field(
        default=default,
        default_factory=default_factory,
        negative_prefix=negative_prefix,
        negative_option=negative_option,
        nargs=nargs,
        type=type,
        action=action,
        **kwargs,
    )


def flags(
    default_factory: Callable[[], list[bool]] | _MISSING_TYPE = MISSING,
    nargs: Literal["*", "+"] | int = "*",
    type: Callable[[str], bool] = str2bool,
    **kwargs,
) -> list[bool]:
    return field(
        default_factory=default_factory,
        nargs=nargs,
        type=type,
        **kwargs,
    )



from __future__ import annotations

from logging import getLogger
from pathlib import Path
from typing import IO

import yaml

from .serializable import D, Serializable

logger = getLogger(__name__)


class YamlSerializable(Serializable):
    """Convenience class, just sets different `load_fn` and `dump_fn` defaults
    for the `dump`, `dumps`, `load`, `loads` methods of `Serializable`.

    Uses the `yaml.safe_load` and `yaml.dump` for loading and dumping.

    Requires the pyyaml package.
    """

    def dump(self, fp: IO[str], dump_fn=yaml.dump, **kwargs) -> None:
        dump_fn(self.to_dict(), fp, **kwargs)

    def dumps(self, dump_fn=yaml.dump, **kwargs) -> str:
        return dump_fn(self.to_dict(), **kwargs)

    @classmethod
    def load(
        cls: type[D],
        path: Path | str | IO[str],
        drop_extra_fields: bool | None = None,
        load_fn=yaml.safe_load,
        **kwargs,
    ) -> D:
        return super().load(path, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def loads(
        cls: type[D],
        s: str,
        drop_extra_fields: bool | None = None,
        load_fn=yaml.safe_load,
        **kwargs,
    ) -> D:
        return super().loads(s, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def _load(
        cls: type[D],
        fp: IO[str],
        drop_extra_fields: bool | None = None,
        load_fn=yaml.safe_load,
        **kwargs,
    ) -> D:
        return super()._load(fp, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)



""" Test for https://github.com/lebrice/SimpleParsing/issues/132 """
from dataclasses import dataclass

from simple_parsing import field

from .conftest import SimpleAttributeTuple
from .testutils import TestSetup


def test_field_with_custom_required_arg_is_optional(
    simple_attribute: SimpleAttributeTuple,
):
    """Test that the `field` function can be used as a work-around for issue 132.

    When passing `required` (or any other of the usual arguments of `parser.add_argument`) to the
    `field` function, they get saved into the `FieldWrapper`, and used to populate the args_dict
    that gets passed to `parser.add_arguments(*field_wrapper.option_strings, **args_dict)`.

    Therefore, using `required=False` as a custom argument is a work-around, while we fix this
    issue.
    """
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class Foo(TestSetup):
        a: some_type = field(default=None, required=False)  # type: ignore

    assert Foo.setup() == Foo(a=None)
    assert Foo.setup(f"--a {passed_value}") == Foo(a=expected_value)


def test_field_with_none_default_is_optional(simple_attribute: SimpleAttributeTuple):
    """Test that when the default value is None, the argument is treated as optional."""
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class Foo(TestSetup):
        a: some_type = None  # type: ignore

    assert Foo.setup() == Foo(a=None)
    assert Foo.setup(f"--a {passed_value}") == Foo(a=expected_value)


def test_dataclass_field_with_none_default_is_optional(
    simple_attribute: SimpleAttributeTuple,
):
    """Test that when the default value is None, the argument is treated as optional."""
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class Foo(TestSetup):
        a: some_type  # type: ignore

    @dataclass
    class Bar(TestSetup):
        foo: Foo = None  # type: ignore

    assert Bar.setup() == Bar(foo=None)  # type: ignore
    assert Bar.setup(f"--a {passed_value}") == Bar(foo=Foo(a=expected_value))



"""A test to make sure that all the example files work without crashing.
(Could be seen as a kind of integration test.)

"""
from __future__ import annotations

import glob
import os
import runpy
import shlex
import sys
from collections import Counter
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Sequence

import pytest

expected = ""


def xfail_param(*args, reason: str):
    return pytest.param(*args, marks=pytest.mark.xfail(reason=reason))


@pytest.fixture
def set_prog_name():
    argv = sys.argv.copy()
    del sys.argv[1:]

    def set_prog(prog_name: str, args: str):
        sys.argv[0] = prog_name
        sys.argv[1:] = shlex.split(args)

    yield set_prog
    sys.argv = argv


@pytest.fixture
def assert_equals_stdout(capsys):
    def strip(string):
        return "".join(string.split())

    def should_equal(expected: str, file_path: str):
        if "optional arguments" in expected and sys.version_info >= (3, 10):
            expected = expected.replace("optional arguments", "options")
        out = capsys.readouterr().out
        # assert strip(out) == strip(expected), file_path
        # assert out == expected, file_path
        out_lines = out.splitlines(keepends=False)
        out_lines = [line.strip() for line in out_lines if line and not line.isspace()]
        expected_lines = expected.splitlines(keepends=False)
        expected_lines = [line.strip() for line in expected_lines if line and not line.isspace()]
        if sys.version_info[:2] >= (3, 9):
            # Ordering of values in a `Namespace` object are different!
            expected = Counter("".join(expected_lines))
            actual = Counter("".join(out_lines))
            actual[" "] = 0
            expected[" "] = 0
            assert actual == expected, (
                out_lines,
                expected_lines,
            )
        else:
            assert out_lines == expected_lines, file_path

    return should_equal


@pytest.mark.parametrize(
    "file_path, args",
    [
        ("examples/subparsers/subparsers_example.py", "train"),
    ],
)
def test_running_example_outputs_expected(
    file_path: str,
    args: str,
    set_prog_name: Callable[[str, str | None], None],
    assert_equals_stdout: Callable[[str, str], None],
):
    # set_prog_name(script, args)
    example_dir = Path(file_path).parent

    file_path = Path(file_path).as_posix()
    module_name = file_path.replace("/", ".")
    if module_name.endswith(".py"):
        module_name = module_name[:-3]
    # programmatically import the example script, which also runs it.
    # (Equivalent to executing "from <module_name> import expected")

    # move to the example directory.
    with temporarily_chdir(example_dir), temporarily_add_args(args):
        resulting_globals = runpy.run_module(
            module_name, init_globals=None, run_name="__main__", alter_sys=True
        )

    if "expected" not in resulting_globals:
        pytest.xfail(reason="Example doesn't have an 'expected' global variable.")
    expected = resulting_globals["expected"]
    assert_equals_stdout(expected, file_path)


@pytest.mark.parametrize(
    "file_path",
    [
        *[
            pytest.param(
                p,
                marks=[
                    pytest.mark.skipif(
                        sys.version_info[:2] == (3, 6),
                        reason="Example uses __future__ annotations feature",
                    ),
                    pytest.mark.xfail(
                        reason="Example has different indentation depending on python version.",
                    ),
                ],
            )
            if p == "examples/subgroups/subgroups_example.py"
            else p
            for p in glob.glob("examples/**/*.py")
            if p
            not in {
                "examples/merging/multiple_lists_example.py",
                "examples/subparsers/subparsers_example.py",
                "examples/serialization/custom_types_example.py",
            }
        ],
    ],
)
def test_running_example_outputs_expected_without_arg(
    file_path: str,
    set_prog_name: Callable[[str, str | None], None],
    assert_equals_stdout: Callable[[str, str], None],
):
    return test_running_example_outputs_expected(file_path, "", set_prog_name, assert_equals_stdout)


@contextmanager
def temporarily_chdir(new_dir: Path):
    """Temporarily navigate to the given directory."""
    start_dir = Path.cwd()
    try:
        os.chdir(new_dir)
        yield
    except OSError:
        raise
    finally:
        os.chdir(start_dir)


@contextmanager
def temporarily_add_args(args: str | Sequence[str]):
    """Temporarily adds the given arguments to sys.argv."""
    if isinstance(args, str):
        args = shlex.split(args)
    start_argv = sys.argv.copy()
    try:
        sys.argv = start_argv + list(args)
        yield
    finally:
        sys.argv = start_argv



from __future__ import annotations

import dataclasses
import functools
import inspect
import shlex
import sys
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Callable, TypeVar

import pytest
from pytest_regressions.file_regression import FileRegressionFixture
from typing_extensions import Annotated

from simple_parsing import ArgumentParser, parse, subgroups
from simple_parsing.wrappers.field_wrapper import ArgumentGenerationMode, NestedMode

from .test_choice import Color
from .testutils import TestSetup, raises_invalid_choice, raises_missing_required_arg

TestClass = TypeVar("TestClass", bound=TestSetup)


@dataclass
class A:
    a: float = 0.0


@dataclass
class B:
    b: str = "bar"


@dataclass
class AB(TestSetup):
    a_or_b: A | B = subgroups({"a": A, "b": B}, default_factory=A)


@dataclass
class C:
    c: bool = False


@dataclass
class D:
    d: int = 0


@dataclass
class E:
    e: bool = False


@dataclass
class F:
    f: str = "f_default"


@dataclass
class G:
    g: int = 0


@dataclass
class H:
    h: bool = False


@dataclass
class CD:
    c_or_d: C | D = subgroups({"c": C, "d": D}, default_factory=C)
    other_c_arg: str = "bob"


@dataclass
class EF:
    e_or_f: E | F = subgroups({"e": E, "f": F}, default_factory=E)


@dataclass
class GH:
    g_or_h: G | H = subgroups({"g": G, "h": H}, default_factory=G)


@dataclass
class ABCD(TestSetup):
    ab_or_cd: AB | CD = subgroups({"ab": AB, "cd": CD}, default_factory=AB)


@dataclass
class EFGH:
    ef_or_gh: EF | GH = subgroups({"ef": EF, "gh": GH}, default_factory=EF)


@dataclass
class ABCDEFGH(TestSetup):
    """Dataclass with three levels of subgroup nesting."""

    abc_or_efgh: ABCD | EFGH = subgroups({"abcd": ABCD, "efgh": EFGH}, default_factory=ABCD)


@dataclass
class MultipleSubgroupsSameLevel(TestSetup):
    a_or_b: A | B = subgroups({"a": A, "b": B}, default_factory=A)
    c_or_d: C | D = subgroups({"c": C, "d": D}, default_factory=D)


@dataclass
class MultipleSubgroupsDifferentLevel(TestSetup):
    ab_or_cd: AB | CD = subgroups({"ab": AB, "cd": CD}, default_factory=CD)
    ef: EF = field(default_factory=EF)


@dataclass
class EnumsAsKeys(TestSetup):
    """Dataclass where the subgroup choices are keys."""

    a_or_b: A | B = subgroups({Color.red: A, Color.blue: B}, default_factory=A)


@pytest.mark.parametrize(
    "dataclass_type, get_help_text_args, should_contain",
    [
        (AB, {}, ["--a_or_b {a,b}", "--a float"]),
        (AB, {}, ["--a_or_b {a,b}       (default: a)", "--a float"]),
        (EnumsAsKeys, {}, ["--a_or_b {Color.red,Color.blue}", "--a float"]),
        (
            MultipleSubgroupsSameLevel,
            {},
            ["--a_or_b {a,b}", "--a float", "--c_or_d {c,d}", "--d int"],
        ),
        (
            MultipleSubgroupsDifferentLevel,
            {},
            ["--ab_or_cd {ab,cd}", "--c_or_d {c,d}", "--e_or_f {e,f}", "--e bool"],
        ),
    ],
)
def test_help_string(
    dataclass_type: type[TestSetup],
    get_help_text_args: dict,
    should_contain: list[str],
):
    """Test that the arguments for the chosen subgroup are shown in the help string."""
    help_text = dataclass_type.get_help_text(*get_help_text_args)
    for expected in should_contain:
        assert expected in help_text


@pytest.mark.parametrize(
    "dataclass_type, args, expected",
    [
        (
            AB,
            "--a_or_b a --a 123",
            AB(a_or_b=A(a=123)),
        ),
        (
            AB,
            "--a_or_b b --b foooo",
            AB(a_or_b=B(b="foooo")),
        ),
        (
            MultipleSubgroupsSameLevel,
            "--a_or_b a --a 123 --d 456",
            MultipleSubgroupsSameLevel(a_or_b=A(a=123), c_or_d=D(d=456)),
        ),
        (
            MultipleSubgroupsSameLevel,
            "--a_or_b b --b foooo",
            MultipleSubgroupsSameLevel(a_or_b=B(b="foooo")),
        ),
        (
            ABCD,
            "--ab_or_cd ab --a_or_b a --a 123",
            ABCD(ab_or_cd=AB(a_or_b=A(a=123))),
        ),
        (
            ABCD,
            "--ab_or_cd cd --c_or_d d --d 456",
            ABCD(ab_or_cd=CD(c_or_d=D(d=456))),
        ),
    ],
)
def test_parse(dataclass_type: type[TestClass], args: str, expected: TestClass):
    assert dataclass_type.setup(args) == expected


def test_subgroup_choice_is_saved_on_namespace():
    """test for https://github.com/lebrice/SimpleParsing/issues/139

    Need to save the chosen subgroup name somewhere on the args.
    """
    parser = ArgumentParser()
    parser.add_arguments(AB, dest="config")

    args = parser.parse_args(shlex.split("--a_or_b b --b foobar"))
    assert args.config == AB(a_or_b=B(b="foobar"))
    assert args.subgroups == {"config.a_or_b": "b"}


@dataclass
class RequiredSubgroup(TestSetup):
    a_or_b: A | B = subgroups({"a": A, "b": B})


def test_required_subgroup():
    """Test when a subgroup doesn't have a default value, and is required."""

    with raises_missing_required_arg():
        assert RequiredSubgroup.setup("")

    assert RequiredSubgroup.setup("--a_or_b b") == RequiredSubgroup(a_or_b=B())


@dataclass
class TwoSubgroupsWithConflict(TestSetup):
    first: AB | CD = subgroups({"ab": AB, "cd": CD}, default_factory=CD)
    second: AB | GH = subgroups({"ab": AB, "gh": GH}, default_factory=GH)


@pytest.mark.parametrize(
    "args_str, expected",
    [
        (
            (
                "--first ab --first.a_or_b a --first.a_or_b.a 111 "
                "--second ab --second.a_or_b a --second.a_or_b.a 234"
            ),
            TwoSubgroupsWithConflict(first=AB(a_or_b=A(a=111)), second=AB(a_or_b=A(a=234))),
        ),
        (
            # TODO: Unsure about this one. Also have to be careful about the abbrev feature of
            # Argparse.
            "--first ab --first.a_or_b a --a 111 --second ab --second.a_or_b b --b arwg",
            TwoSubgroupsWithConflict(first=AB(a_or_b=A(a=111)), second=AB(a_or_b=B(b="arwg"))),
        ),
    ],
)
def test_two_subgroups_with_conflict(args_str: str, expected: TwoSubgroupsWithConflict):
    assert TwoSubgroupsWithConflict.setup(args_str) == expected


def test_subgroups_with_key_default() -> None:

    with pytest.raises(ValueError):
        subgroups({"a": A, "b": B}, default_factory="a")

    @dataclass
    class Foo(TestSetup):
        a_or_b: A | B = subgroups({"a": A, "b": B}, default="a")

    assert Foo.setup() == Foo(a_or_b=A())
    assert Foo.setup("--a_or_b a --a 445") == Foo(a_or_b=A(a=445))
    assert Foo.setup("--a_or_b b") == Foo(a_or_b=B())
    assert Foo.setup("--a_or_b b --b zodiak") == Foo(a_or_b=B(b="zodiak"))


# IDEA: Make it possible to use a default factory that is a partial for a function, if that
# function has a return annotation.
# def some_a_factory(a: int = -1) -> A:
#     return A(a=a)


def test_subgroup_default_needs_to_be_key_in_dict():
    with pytest.raises(ValueError, match="default must be a key in the subgroups dict"):
        _ = subgroups({"a": B, "aa": A}, default="b")


def test_subgroup_default_factory_needs_to_be_value_in_dict():
    with pytest.raises(ValueError, match="`default_factory` must be a value in the subgroups dict"):
        _ = subgroups({"a": B, "aa": A}, default_factory=C)


def test_lambdas_dont_return_same_instance():
    """Slightly unrelated, but I just want to check if lambda expressions return the same object
    instance when a default factory looks like `lambda: A()`. If so, then I won't encourage this.
    """

    @dataclass
    class A:
        a: int = 123

    @dataclass
    class Config(TestSetup):
        a_config: A = dataclasses.field(default_factory=lambda: A())

    assert Config().a_config is not Config().a_config


def test_partials_new_args_overwrite_set_values():
    """Double-check that functools.partial overwrites the keywords that are stored when it is
    created with the ones that are passed when calling it.
    """
    # just to avoid the test passing if I were to hard-code the same value as the default by
    # accident.
    default_a = A().a
    CustomA = functools.partial(A, a=default_a + 1)
    assert CustomA() == A(a=default_a + 1)
    assert CustomA(a=default_a * 34) == A(a=default_a * 34)


def test_defaults_from_partial():
    @dataclass
    class Foo(TestSetup):
        a_or_b: A | B = subgroups(
            {
                "a": A,
                "a_1.23": functools.partial(A, a=1.23),
                "a_4.56": functools.partial(A, a=4.56),
                "b": B,
                "b_bob": functools.partial(B, b="bob"),
            },
            default="a",
        )

    # TODO: The issue is that the default value from the field is being set in the dict of
    # constructor arguments, and even if the dataclass_fn is a partial(A, a=1.23), since it's being
    # called like `dataclass_fn(**{"a": 0.0})` (from the field default), then the value of `a` is
    # overwritten with the default value from the field. I think the solution is either:
    # 1. Not populate the constructor arguments with the value for this field;
    # 2. Change the default value to be the one from the partial, instead of the one from the
    #    field. The partial would then be called with the same value.
    # I think 1. makes more sense. For fields that aren't required (have a default value), then the
    # constructor_arguments dict doesn't need to contain the default values. Calling the
    # dataclass_fn (which is almost always just the dataclass type itself) will use the default
    # values from the fields.

    # NOTE: Now I've changed my mind. Option 2 might be easiest for now, since I still don't have a
    # complete grasp of how argparse works internally, and I feel like I could easily make it work
    # quickly.
    assert Foo.setup("--a_or_b a_4.56") == Foo(a_or_b=A(a=4.56))


def test_subgroups_with_partials():
    @dataclass
    class Foo(TestSetup):
        a_or_b: A | B = subgroups(
            {
                "a": A,
                "a_1.23": functools.partial(A, a=1.23),
                "a_4.56": functools.partial(A, a=4.56),
                "b": B,
                "b_bob": functools.partial(B, b="bob"),
            },
            default="a",
        )

    assert Foo.setup() == Foo() == Foo(a_or_b=A())
    assert Foo.setup("--a_or_b a") == Foo(a_or_b=A())
    assert Foo.setup("--a_or_b a --a=6.66") == Foo(a_or_b=A(a=6.66))
    assert Foo.setup("--a_or_b a_1.23") == Foo(a_or_b=A(a=1.23))
    assert Foo.setup("--a_or_b a_4.56") == Foo(a_or_b=A(a=4.56))
    assert Foo.setup("--a_or_b b") == Foo(a_or_b=B())
    assert Foo.setup("--a_or_b b --b fooo") == Foo(a_or_b=B(b="fooo"))
    assert Foo.setup("--a_or_b b_bob") == Foo(a_or_b=B(b="bob"))

    with raises_invalid_choice():
        Foo.setup("--a_or_b c")


@pytest.mark.xfail(strict=True, reason="I'm not sure this is a good idea anymore.")
def test_subgroups_with_functions():
    def make_b(b: str = "default from make_b") -> B:
        return B(b=b)

    @dataclass
    class Foo(TestSetup):
        a_or_b: A | B = subgroups(
            {
                "a": A,
                "b": B,
                "make_b": make_b,
            },
            default="a",
        )

    # TODO: decide if this is correct. In this case here, since the default value for `b` is the
    # same as passed (i.e. arg isn't used), then the constructor args dict doesn't have an entry
    # for `b`, and the default value from the function definition is used.
    # This also means that `b` can't be a required argument in this function, since it won't be
    # passed if it isn't different from the default value..
    # TODO: Now that I think about it, it might make more sense to have the default values in the
    # dictionary of constructor arguments, right? Not 100% sure.

    assert Foo.setup("--a_or_b make_b") == Foo(a_or_b=B(b="default from make_b"))
    assert Foo.setup("--a_or_b make_b") == Foo(a_or_b=B(b="default from make_b"))
    assert Foo.setup("--a_or_b make_b --b foo") == Foo(a_or_b=B(b="foo"))


def test_subgroup_functions_receive_all_fields():
    """TODO: Decide how we want to go about this.
    Either the functions receive all the fields (the default values), or only the ones that are set
    (harder to implement).
    """

    @dataclass
    class Obj:
        a: float = 0.0
        b: str = "default from field"

    def make_obj(**kwargs) -> Obj:
        assert kwargs == {"a": 0.0, "b": "foo"}  # first case (current): receives all fields
        # assert kwargs == {"b": "foo"}  # second case: receive only set fields.
        return Obj(**kwargs)

    @dataclass
    class Foo(TestSetup):
        a_or_b: Obj = subgroups(
            {
                "make_obj": make_obj,
            },
            default_factory=make_obj,
        )

    Foo.setup("--a_or_b make_obj --b foo")


lambdas_arent_supported_yet = functools.partial(
    pytest.param,
    marks=pytest.mark.xfail(
        strict=True,
        raises=NotImplementedError,
        reason="Lambda expressions aren't allowed in the subgroup dict or default_factory at the moment.",
    ),
)


@pytest.mark.parametrize(
    "a_factory, b_factory",
    [
        (partial(A), partial(B)),
        (partial(A, a=321), partial(B, b="foobar")),
        lambdas_arent_supported_yet(lambda: A(), lambda: B()),
        lambdas_arent_supported_yet(lambda: A(a=123), lambda: B(b="foooo")),
    ],
)
def test_other_default_factories(a_factory: Callable[[], A], b_factory: Callable[[], B]):
    """Test using other kinds of default factories (i.e. functools.partial or lambda expressions)"""

    @dataclass
    class Foo(TestSetup):
        a_or_b: A | B = subgroups({"a": a_factory, "b": b_factory}, default="a")

    assert Foo.setup() == Foo(a_or_b=a_factory())
    assert Foo.setup("--a_or_b a --a 445") == Foo(a_or_b=A(a=445))
    assert Foo.setup("--a_or_b b") == Foo(a_or_b=b_factory())


@pytest.mark.parametrize(
    "a_factory, b_factory",
    [
        (partial(A, a=321), partial(B, b="foobar")),
        (partial(partial(A, a=111), a=321), partial(partial(B), b="foobar")),
        (partial(partial(A, a=111)), partial(partial(B, b="foobar"))),
        lambdas_arent_supported_yet(lambda: A(a=123), lambda: B(b="foooo")),
    ],
)
def test_help_string_displays_default_factory_arguments(
    a_factory: Callable[[], A], b_factory: Callable[[], B]
):
    """The help string should be basically the same when using a `partial` or a lambda that returns
    a dataclass, as using just the class itself.

    When using `functools.partial` or lambda expressions, we'd ideally also like the help text to
    show the field values from inside the `partial` or lambda, if possible.
    """
    # NOTE: Here we need to return just A() and B() with these default factories, so the defaults
    # for the fields are the same
    @dataclass
    class Foo(TestSetup):
        a_or_b: A | B = subgroups({"a": a_factory, "b": b_factory}, default="a")

    a_default_from_factory = a_factory().a
    b_default_from_factory = b_factory().b
    a_default_from_field = A().a
    b_default_from_field = B().b

    # Just to make sure that we're using a different value for the default.
    assert a_default_from_field != a_default_from_factory
    assert b_default_from_field != b_default_from_factory

    # Check that it doesn't use the default value from the field in the help text.
    assert f"--a float (default: {a_default_from_field})" not in Foo.get_help_text("")
    assert f"--a float (default: {a_default_from_field})" not in Foo.get_help_text("--a_or_b a")
    assert f"--b str (default: {b_default_from_field})" not in Foo.get_help_text("--a_or_b b")

    # Check that it uses the default value from the factory in the help text.
    assert f"--a float  (default: {a_default_from_factory})" in Foo.get_help_text("")
    assert f"--a float  (default: {a_default_from_factory})" in Foo.get_help_text("--a_or_b a")
    assert f"--b str  (default: {b_default_from_factory})" in Foo.get_help_text("--a_or_b b")


def test_factory_is_only_called_once():
    class_constructor_calls = 0
    partial_calls = 0

    class _partial(partial):
        def __call__(self, *args, **kwargs):
            nonlocal partial_calls
            partial_calls += 1
            return super().__call__(*args, **kwargs)

    @dataclass
    class SomeObj:
        a: int = 0

        def __post_init__(self):
            nonlocal class_constructor_calls
            class_constructor_calls += 1

    @dataclass
    class Config(TestSetup):
        obj: SomeObj = subgroups(
            {
                "a": SomeObj,
                "b": _partial(SomeObj, a=123),
            },
            default_factory=SomeObj,
        )

    assert class_constructor_calls == 0
    Config()
    assert class_constructor_calls == 1

    config = Config.setup("")
    assert class_constructor_calls == 2
    assert config.obj.a == 0

    config = Config.setup("--obj a --a 321")
    assert class_constructor_calls == 3
    assert config == Config(obj=SomeObj(a=321))
    assert class_constructor_calls == 4

    assert partial_calls == 0
    config = Config.setup("--obj b")
    assert partial_calls == 1
    assert class_constructor_calls == 5
    assert config.obj.a == 123


@pytest.mark.xfail(strict=True, reason="Not implemented yet. Remove this once it is.")
def test_all_subgroups_are_in_help_string():
    @dataclass
    class Foo(TestSetup):
        a_or_b: A | B = subgroups({"a": A, "b": B}, default_factory=B)

    help_text = Foo.get_help_text()

    assert "-a float, --a float  (default: 0.0)" in help_text
    assert "-b str, --b str  (default: bar)" in help_text


def test_typing_of_subgroups_function() -> None:
    """TODO: There should be a typing errors here. Could we check for it programmatically?"""
    # note: This should raise an error with the type checker:
    bob: A = subgroups({"a": A, "b": B})  # noqa: F841
    # reveal_type(bob)

    with pytest.raises(ValueError):
        other: A | B = subgroups({"a": A, "b": B}, default_factory=C)  # noqa: F841
    # reveal_type(other)


@dataclass
class ModelConfig:
    ...


@dataclass
class ModelAConfig(ModelConfig):
    lr: float = 3e-4
    optimizer: str = "Adam"
    betas: tuple[float, float] = (0.9, 0.999)


@dataclass
class ModelBConfig(ModelConfig):
    lr: float = 1e-3
    optimizer: str = "SGD"
    momentum: float = 1.234


@dataclass
class Config(TestSetup):

    # Which model to use
    model: ModelConfig = subgroups(
        {"model_a": ModelAConfig, "model_b": ModelBConfig},
        default_factory=ModelAConfig,
    )


def test_destination_substring_of_other_destination_issue191():
    """Test for https://github.com/lebrice/SimpleParsing/issues/191"""

    parser = ArgumentParser()
    parser.add_arguments(Config, dest="config")
    parser.add_arguments(Config, dest="config2")  # this produces and exception
    # parser.add_arguments(Config, dest="something") # this works as expected
    args = parser.parse_args("")

    config: Config = args.config
    assert config.model == ModelAConfig()


def test_subgroup_partial_with_nested_field():
    """Test the case where a subgroup has a nested dataclass field, and that nested field is an
    argument to the partial."""

    @dataclass
    class Obj:
        a: int = 0

    @dataclass
    class Foo:
        obj: Obj = field(default_factory=Obj)

    @dataclass
    class Config(TestSetup):
        foo: Foo = subgroups(
            {
                "simple": Foo,
                "default": partial(Foo, obj=Obj(a=123)),  # bad idea!
            },
            default="default",
        )

    first_config = Config()
    first_object = first_config.foo.obj
    second_config = Config()
    # Bad idea! This is reusing the dataclass instance!
    # TODO: Do we want to explicitly disallow this?
    assert second_config.foo.obj is first_object

    assert Config.setup("").foo.obj.a == 123
    assert Config.setup("--foo=default").foo.obj.a == 123
    assert Config.setup("--foo=simple").foo.obj.a == 0
    assert "--a int       (default: 123)" in Config.get_help_text()


class Model:
    def __init__(self, num_layers: int = 3, hidden_dim: int = 64):
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim


@dataclass
class CustomAnnotation:
    default_factory: Callable[[], Model]


SmallModel = Annotated[
    Model, CustomAnnotation(default_factory=partial(Model, num_layers=1, hidden_dim=32))
]
BigModel = Annotated[
    Model, CustomAnnotation(default_factory=partial(Model, num_layers=12, hidden_dim=128))
]


@pytest.mark.xfail(strict=True, reason="Annotated isn't supported as an option yet.")
def test_annotated_as_subgroups():
    """Test using the new Annotated type to create variants of the same dataclass."""

    @dataclasses.dataclass
    class Config(TestSetup):
        model: Model = subgroups({"small": SmallModel, "big": BigModel}, default_factory=SmallModel)

    assert Config.setup().model == SmallModel()
    # Hopefully this illustrates why Annotated aren't exactly great:
    # At runtime, they are basically the same as the original dataclass when called.
    assert SmallModel() == Model()
    assert BigModel() == Model()

    assert Config.setup("--model small").model == Model(num_layers=1, hidden_dim=32)
    assert Config.setup("--model big").model == Model(num_layers=12, hidden_dim=128)
    assert Config.setup("--num_layers 123").model == Model(num_layers=123, hidden_dim=32)


@dataclasses.dataclass(frozen=True)
class FrozenConfig:
    a: int = 1
    b: str = "bob"


def test_subgroups_doesnt_support_nonfrozen_instances():
    with pytest.raises(
        ValueError,
        match="'default' can either be a key of the subgroups dict or a hashable",
    ):
        _ = subgroups({"a": A, "b": B}, default=A(a=1.0))

    with pytest.raises(
        ValueError,
        match="'default' can either be a key of the subgroups dict or a hashable",
    ):
        _ = subgroups({"a": A(a=1.0), "b": B}, default=A(a=1.0))


odd = FrozenConfig(a=1, b="odd")
even = FrozenConfig(a=2, b="even")


@dataclasses.dataclass
class ConfigWithFrozen(TestSetup):
    conf: FrozenConfig = subgroups({"odd": odd, "even": even}, default=odd)


@pytest.mark.parametrize(
    ("command", "expected"),
    [
        ("", ConfigWithFrozen(odd)),
        ("--conf=odd", ConfigWithFrozen(odd)),
        ("--conf=even", ConfigWithFrozen(even)),
        ("--conf=odd --a 123", ConfigWithFrozen(dataclasses.replace(odd, a=123))),
        ("--conf=even --a 100", ConfigWithFrozen(dataclasses.replace(even, a=100))),
    ],
)
def test_subgroups_supports_frozen_instances(command: str, expected: ConfigWithFrozen):
    assert ConfigWithFrozen.setup(command) == expected


@pytest.mark.parametrize(
    ("dataclass_type", "command"),
    [
        (Config, "--help"),
        (Config, "--model=model_a --help"),
        (Config, "--model=model_b --help"),
        (ConfigWithFrozen, "--help"),
        (ConfigWithFrozen, "--conf=odd --help"),
        (ConfigWithFrozen, "--conf=even --help"),
        (ConfigWithFrozen, "--conf=odd --a 123 --help"),
        (ConfigWithFrozen, "--conf=even --a 100 --help"),
    ],
)
def test_help(
    dataclass_type: type[TestSetup],
    command: str,
    file_regression: FileRegressionFixture,
    request: pytest.FixtureRequest,
):
    if sys.version_info[:2] != (3, 11):
        pytest.skip("The regression check is only ran with Python 3.11")

    here = Path(__file__).relative_to(Path.cwd())
    file_regression.check(
        f"""\
# Regression file for [this test]({here}:{inspect.getsourcelines(test_help)[1]})

Given Source code:

```python
{''.join(inspect.getsourcelines(dataclass_type)[0])}
```

and command: {command!r}

We expect to get:

```console
{dataclass_type.get_help_text(command, prog="pytest")}
```
""",
        basename=request.node.name,
        extension=".md",
    )


# def test_subgroups_with_partials():
#     class Model:
#         def __init__(self, num_layers: int = 3, hidden_dim: int = 64):
#             self.num_layers = num_layers
#             self.hidden_dim = hidden_dim

#     _ModelConfig: type[Partial[Model]] = config_dataclass_for(Model)

#     ModelConfig = _ModelConfig()
#     SmallModel = _ModelConfig(num_layers=1, hidden_dim=32)
#     BigModel = _ModelConfig(num_layers=32, hidden_dim=128)

#     @dataclasses.dataclass
#     class Config(TestSetup):
#         model: Model = subgroups({"small": SmallModel, "big": BigModel}, default_factory=SmallModel)

#     assert Config.setup().model == SmallModel()
#     # Hopefully this illustrates why Annotated aren't exactly great:
#     # At runtime, they are basically the same as the original dataclass when called.
#     assert SmallModel() != Model()
#     assert SmallModel() == Model(num_layers=1, hidden_dim=32)
#     assert BigModel() != Model()
#     assert BigModel() == Model(num_layers=32, hidden_dim=128)

#     assert Config.setup("--model small").model == SmallModel()
#     assert Config.setup("--model big").model == BigModel()
#     assert Config.setup("--num_layers 123").model == Model(num_layers=123, hidden_dim=32)


@pytest.mark.parametrize("frozen", [True, False])
def test_nested_subgroups(frozen: bool):
    """Assert that #160 is fixed: https://github.com/lebrice/SimpleParsing/issues/160"""

    @dataclass(frozen=frozen)
    class FooConfig:
        ...

    @dataclass(frozen=frozen)
    class BarConfig:
        foo: FooConfig

    @dataclass(frozen=frozen)
    class FooAConfig(FooConfig):
        foo_param_a: float = 0.0

    @dataclass(frozen=frozen)
    class FooBConfig(FooConfig):
        foo_param_b: str = "foo_b"

    @dataclass(frozen=frozen)
    class Bar1Config(BarConfig):
        foo: FooConfig = subgroups(
            {"foo_a": FooAConfig, "foo_b": FooBConfig},
            default_factory=FooAConfig,
        )

    @dataclass(frozen=frozen)
    class Bar2Config(BarConfig):
        foo: FooConfig = subgroups(
            {"foo_a": FooAConfig, "foo_b": FooBConfig},
            default_factory=FooBConfig,
        )

    @dataclass(frozen=frozen)
    class Config(TestSetup):
        bar: Bar1Config | Bar2Config = subgroups(
            {"bar_1": Bar1Config, "bar_2": Bar2Config},
            default_factory=Bar2Config,
        )

    assert Config.setup("") == Config(bar=Bar2Config(foo=FooBConfig()))
    assert Config.setup("--bar=bar_1 --foo=foo_a") == Config(bar=Bar1Config(foo=FooAConfig()))


@dataclass
class ModelConfig:
    ...


@dataclass
class DatasetConfig:
    ...


@dataclass
class ModelAConfig(ModelConfig):
    lr: float = 3e-4
    optimizer: str = "Adam"
    betas: tuple[float, float] = (0.9, 0.999)


@dataclass
class ModelBConfig(ModelConfig):
    lr: float = 1e-3
    optimizer: str = "SGD"
    momentum: float = 1.234


@dataclass
class Dataset1Config(DatasetConfig):
    data_dir: str | Path = "data/foo"
    foo: bool = False


@dataclass
class Dataset2Config(DatasetConfig):
    data_dir: str | Path = "data/bar"
    bar: float = 1.2


@dataclass
class Config(TestSetup):

    # Which model to use
    model: ModelConfig = subgroups(
        {"model_a": ModelAConfig, "model_b": ModelBConfig},
        default_factory=ModelAConfig,
    )

    # Which dataset to use
    dataset: DatasetConfig = subgroups(
        {"dataset_1": Dataset1Config, "dataset_2": Dataset2Config},
        default_factory=Dataset2Config,
    )


def _parse_config(args: str) -> Config:
    return parse(
        Config,
        args=args,
        argument_generation_mode=ArgumentGenerationMode.NESTED,
        nested_mode=NestedMode.WITHOUT_ROOT,
    )


def test_ordering_of_args_doesnt_matter():
    """Test to confirm that #160 is fixed:"""

    # $ python issue.py --model model_a --model.lr 1e-2
    assert _parse_config(args="--model model_a --model.lr 1e-2") == Config(
        model=ModelAConfig(lr=0.01, optimizer="Adam", betas=(0.9, 0.999)),
        dataset=Dataset2Config(data_dir="data/bar", bar=1.2),
    )

    # I was expecting this to work given that both model configs have `lr` attribute
    # $ python issue.py --model.lr 1e-2.
    assert _parse_config(args="--model.lr 1e-2") == Config(
        model=ModelAConfig(lr=1e-2, optimizer="Adam", betas=(0.9, 0.999)),
        dataset=Dataset2Config(data_dir="data/bar", bar=1.2),
    )

    # $ python issue.py --model model_a --model.betas 0. 1.
    assert _parse_config(args="--model model_a --model.betas 0. 1.") == Config(
        model=ModelAConfig(lr=0.0003, optimizer="Adam", betas=(0.0, 1.0)),
        dataset=Dataset2Config(data_dir="data/bar", bar=1.2),
    )

    # % ModelA being the default, I was expecting this two work
    # $ python issue.py --model.betas 0. 1.
    assert _parse_config(args="--model.betas 0. 1.") == Config(
        model=ModelAConfig(lr=0.0003, optimizer="Adam", betas=(0.0, 1.0)),
        dataset=Dataset2Config(data_dir="data/bar", bar=1.2),
    )



from dataclasses import dataclass
from typing import Sequence, Union

import pytest

from .hparam import categorical, log_uniform, uniform
from .hyperparameters import HyperParameters

numpy_installed = False
try:
    import numpy as np

    numpy_installed = True
except ImportError:
    pass


@dataclass
class A(HyperParameters):
    learning_rate: float = uniform(0.0, 1.0)


@dataclass
class B(A):
    momentum: float = uniform(0.0, 1.0)


@pytest.mark.skipif(not numpy_installed, reason="Test requires numpy.")
def test_to_array():
    b: B = B.sample()
    array = b.to_array()
    assert np.isclose(array[0], b.learning_rate)
    assert np.isclose(array[1], b.momentum)


@pytest.mark.skipif(not numpy_installed, reason="Test requires numpy.")
def test_from_array():
    array = np.arange(2)
    b: B = B.from_array(array)
    assert b.learning_rate == 0.0
    assert b.momentum == 1.0


@dataclass
class C(HyperParameters):
    lr: float = uniform(0.0, 1.0)
    momentum: float = uniform(0.0, 1.0)


def test_clip_within_bounds():
    """Test to make sure that the `clip_within_bounds` actually restricts the
    values of the HyperParameters to be within the bounds.
    """
    # valid range for learning_rate is (0 - 1].
    a = A(learning_rate=123)
    assert a.learning_rate == 123
    a = a.clip_within_bounds()
    assert a.learning_rate == 1.0

    b = B(learning_rate=0.5, momentum=456)
    assert b.clip_within_bounds() == B(learning_rate=0.5, momentum=1)

    # Test that the types are maintained.
    @dataclass
    class C(HyperParameters):
        a: int = uniform(123, 456, discrete=True)
        b: float = log_uniform(4.56, 123.456)
        c: str = categorical("foo", "bar", "baz")

    with pytest.raises(TypeError):
        _ = C()

    with pytest.raises(TypeError):
        _ = C(a=123)

    with pytest.raises(TypeError):
        _ = C(b=4.56)

    with pytest.raises(TypeError):
        _ = C(c="bar")

    # TODO: IDEA: how about we actually do some post-processing to always clip stuff
    # between bounds?
    # Check that it doesn't change anything if the values are within the range.
    assert C(a=1000, b=1.23, c="bar").clip_within_bounds() == C(456, 4.56, "bar")
    assert C(a=-1.234, b=1000, c="foo").clip_within_bounds() == C(123, 123.456, "foo")


def test_strict_bounds():
    """When creating a class and using a hparam field with `strict=True`, the values
    will be restricted to be within the given bounds.
    """

    @dataclass
    class C(HyperParameters):
        a: int = uniform(0, 1, strict=True)
        b: float = log_uniform(4.56, 123.456)
        c: str = categorical("foo", "bar", "baz", default="foo", strict=True)

    # valid range for a is [0, 1).
    with pytest.raises(
        ValueError,
        match="Field 'a' got value 123, which is outside of the defined prior",
    ):
        _ = C(a=123, b=10, c="foo")

    with pytest.raises(
        ValueError,
        match="Field 'c' got value 'yolo', which is outside of the defined prior",
    ):
        _ = C(a=0.5, b=0.1, c="yolo")

    # should NOT raise an error, since the field `b` isn't strict.
    _ = C(a=0.1, b=-1.26, c="bar")


def test_nesting():
    @dataclass
    class Child(HyperParameters):
        foo: int = uniform(0, 10, default=5)

    from simple_parsing import mutable_field

    @dataclass
    class Parent(HyperParameters):
        child_a: Child = mutable_field(Child, foo=3)

    parent = Parent.sample()
    assert isinstance(parent, Parent)
    assert isinstance(parent.child_a, Child)


def test_choice_field():
    @dataclass
    class Child(HyperParameters):
        hparam: float = categorical(
            {
                "a": 1.23,
                "b": 4.56,
                "c": 7.89,
            },
            default=1.23,
        )

    bob = Child()
    assert bob.hparam == 1.23

    bob = Child.sample()
    assert bob.hparam in {1.23, 4.56, 7.89}
    assert Child.get_orion_space_dict() == {"hparam": "choices(['a', 'b', 'c'], default_value='a')"}


def test_choice_field_with_values_of_a_weird_type():
    @dataclass
    class Bob(HyperParameters):
        hparam_type: float = categorical(
            {
                "a": A,
                "b": B,
                "c": C,
            },
            probabilities={
                "a": 0.1,
                "b": 0.2,
                "c": 0.7,
            },
            default=B,
        )

    bob = Bob()
    assert bob.hparam_type == B

    bob = Bob.sample()
    assert bob.hparam_type in {A, B, C}
    assert Bob.get_orion_space_dict() == {
        "hparam_type": "choices({'a': 0.1, 'b': 0.2, 'c': 0.7}, default_value='b')"
    }


@pytest.mark.xfail(
    reason="TODO: it isn't trivial how to fix this, without having to rework the "
    "from_dict from simple-parsing."
)
def test_replace_int_or_float_preserves_type():
    @dataclass
    class A(HyperParameters):
        # How much of training dataset to check (floats = percent, int = num_batches)
        limit_train_batches: Union[int, float] = 1.0
        # How much of validation dataset to check (floats = percent, int = num_batches)
        limit_val_batches: Union[int, float] = 1.0
        # How much of test dataset to check (floats = percent, int = num_batches)
        limit_test_batches: Union[int, float] = 1.0

    a = A()
    assert isinstance(a.limit_test_batches, float)
    b = a.replace(limit_train_batches=0.5)
    assert isinstance(b.limit_test_batches, float)


try:
    from orion.core.io.space_builder import SpaceBuilder

    orion_installed = True
except ImportError:
    orion_installed = False


@dataclass
class Foo(HyperParameters):
    x: Sequence[int] = uniform(0, 10, default=(5, 5), shape=2)
    y: Sequence[int] = uniform(0, 10, default=(5, 5, 5), shape=3)
    z: Sequence[int] = uniform(0, 10, default=2, shape=5)


def test_priors_with_shape():

    foo = Foo()
    assert foo.x == (5, 5)
    assert foo.y == (5, 5, 5)
    assert foo.z == (2, 2, 2, 2, 2)

    foo = Foo.sample()
    assert len(foo.x) == 2
    assert len(foo.y) == 3
    assert len(foo.z) == 5


@pytest.mark.xfail(reason="Need to update this since spaces give Trials, not points.")
@pytest.mark.skipif(not numpy_installed, reason="Test requires numpy.")
@pytest.mark.skipif(not orion_installed, reason="Test requires Orion.")
def test_contains():
    # TODO: Add a convenience method for creating a Space object from Orion directly.
    foo = Foo(x=(2, 3), y=(1, 2, 3), z=(1, 2, 3, 4, 5))
    space_builder = SpaceBuilder()
    space_config = Foo.get_orion_space_dict()
    space = space_builder.build(space_config)
    assert foo.to_array() in space


def test_field_types():
    @dataclass
    class C(HyperParameters):
        a: int = uniform(123, 456)
        b: float = uniform(4.56, 123.456, default=123)
        c: float = uniform(4.56, 123.456, default=10, discrete=True)
        d: float = uniform(4.56, 123.456, default=10, discrete=False)
        e: float = uniform(4.56, 123.456, default=10, discrete=False, shape=2)
        f: float = uniform(10, 100, default=20, shape=2)

    cs = [C.sample() for _ in range(100)]
    assert C.get_priors()["a"].discrete is True
    assert all(isinstance(c.a, int) for c in cs)

    assert C.get_priors()["b"].discrete is False
    assert all(isinstance(c.b, float) for c in cs)

    assert C.get_priors()["c"].discrete is True
    assert all(isinstance(c.c, int) for c in cs)

    assert C.get_priors()["d"].discrete is False
    assert all(isinstance(c.d, float) for c in cs)

    assert C.get_priors()["e"].discrete is False
    assert all(all(isinstance(v, float) for v in c.e) for c in cs)
    assert C.get_priors()["f"].discrete is True

    if numpy_installed:
        assert all(c.f.dtype == int for c in cs)
    else:
        assert all(all(isinstance(v, int) for v in c.f) for c in cs)
